(*****************************************************************************)
(* Auxiliary functions for "Experiments 1"                                   *)
(*                                                                           *)
(* Adapted from implementation by Hanne Riis Nielson.                        *)
(*                                              23.09.2003. Mikael Buchholtz *)
(*****************************************************************************)


structure AuxEx1 =

struct

local 
    open Lysa
    open MLysa2Lysa


    (***********************************************************************) 
    (* Auxiliary functions for pretty printing                             *)
    (***********************************************************************) 

    fun int2str' a           = if a < 0
			       then "\\hbox{-}"^Int.toString (~a)
                               else Int.toString a

    fun prSub []                 = ""
      | prSub (n::ns)            = n ^ (prSub ns)
				    
    fun prSymbol (n,[])          = n
      | prSymbol (n,ns)          = n ^ "_{" ^ (prSub ns) ^ "}"

    and prSymbols []             = ""
      | prSymbols [x]            = prSymbol x
      | prSymbols (x::xs)        = prSymbol x ^ " ," ^ prSymbols xs

    fun prForm (ORIG ls)         = "\\ORIG{" ^ prSymbols ls ^ "}"
      | prForm (DEST ls)         = "\\DEST{" ^ prSymbols ls ^ "}"

    fun prLs ((i,x)::l::ls)      = i ^ "=" ^ int2str' x ^ "," ^ prLs (l::ls)
      | prLs [(i,x)]             = i ^ "=" ^ int2str' x
      | prLs []                  = ""


    fun prTerm (NAME  n)         = prSymbol n
      | prTerm (NAMEP (n,is))    = prSymbol ("{"^n^"}^{+}",is)
      | prTerm (NAMEM (n,is))    = prSymbol ("{"^n^"}^{-}",is)
      | prTerm (VAR n)           = prSymbol n
      | prTerm (ENC(ts,t,l,NONE)) 
                                 = "\\ENCO{" ^ prTerms ts ^ "}{" 
                                             ^ prTerm t ^ "}{"
                                             ^ prSymbol l ^ "}"
      | prTerm (ENC(ts,t,l,SOME F)) 
	                          = "\\ENC{" ^ prTerms ts ^ "}{" 
                                             ^ prTerm t ^ "}{"
                                             ^ prSymbol l ^ "}{"
                                             ^ prForm F ^ "}"
      | prTerm (AENC(ts,t,l,NONE)) 
                                 = "\\AENCO{" ^ prTerms ts ^ "}{" 
                                              ^ prTerm t ^ "}{"
                                              ^ prSymbol l ^ "}"
      | prTerm (AENC(ts,t,l,SOME F)) 
	                          = "\\AENC{" ^ prTerms ts ^ "}{" 
                                              ^ prTerm t ^ "}{"
                                              ^ prSymbol l ^ "}{"
                                              ^ prForm F ^ "}"

    and prTerms (t::t'::ts)      = prTerm t ^ " ," ^ prTerms (t'::ts)
      | prTerms [t]              = prTerm t
      | prTerms []               = ""

    
    fun prProc NIL               = "\\NIL"
      | prProc (OUT(ts,NIL))     = "\\OUTO{" ^ prTerms ts ^ "}" 
      | prProc (OUT(ts,P))       = "\\OUT{"  ^ prTerms ts ^ "}\n{" 
                                    ^ prProc P ^ "}"
      | prProc (INP(ts,vs,NIL))  = "\\INPO{" ^ prTerms ts ^ "}{" 
                                   ^ prTerms vs ^ "}"
      | prProc (INP(ts,vs,P))    = "\\INP{"  ^ prTerms ts ^ "}{" 
                                   ^ prTerms vs ^ "}\n{" 
                                   ^ prProc P ^ "}" 
      | prProc (PAR(Ps))         = "\\begin{array}[t]{ll} & " ^ 
				   prProcs Ps ^
				   "\\end{array}"
      | prProc (NEW(t,P))        = "\\NEW{" ^ prTerm t ^ "}\n{"
                                   ^ prProc P^"}"
      | prProc (ANEW(t,P))       = "\\ANEW{" ^ prTerm t ^ "}\n{"
                                   ^ prProc P^"}"
      | prProc (BANG P)          = "\\BANG{" ^ prProc P ^ "}"
      | prProc (DEC(t,ts,vs,t',l,NONE,NIL)) 
                                 = "\\DECOO{" ^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
					      ^ prTerms vs ^ "}{"
					      ^ prTerm t' ^ "}{"
					      ^ prSymbol l ^ "}"
      | prProc (DEC(t,ts,vs,t',l,NONE,P)) 
                                 = "\\DECOP{" ^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
                                              ^ prTerms vs ^ "}{"
                                              ^ prTerm t' ^ "}{"
                                              ^ prSymbol l ^ "}\n{"
                                              ^ prProc P ^ "}"
      | prProc (DEC(t,ts,vs,t',l,SOME F,NIL)) 
                                 = "\\DECFO{" ^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
                                              ^ prTerms vs ^ "}{"
                                              ^ prTerm t' ^ "}{"
                                              ^ prSymbol l ^ "}{"
                                              ^ prForm F ^ "}"
      | prProc (DEC(t,ts,vs,t',l,SOME F,P)) 
                                 = "\\DECFP{" ^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
                                              ^ prTerms vs ^ "}{"
                                              ^ prTerm t' ^ "}{"
                                              ^ prSymbol l ^ "}{"
                                              ^ prForm F ^ "}\n{" 
                                              ^ prProc P ^ "}"
      | prProc (ADEC(t,ts,vs,t',l,NONE,NIL)) 
                                 = "\\ADECOO{"^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
					      ^ prTerms vs ^ "}{"
					      ^ prTerm t' ^ "}{"
					      ^ prSymbol l ^ "}"
      | prProc (ADEC(t,ts,vs,t',l,NONE,P)) 
                                 = "\\ADECOP{"^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
                                              ^ prTerms vs ^ "}{"
                                              ^ prTerm t' ^ "}{"
                                              ^ prSymbol l ^ "}\n{"
                                              ^ prProc P ^ "}"
      | prProc (ADEC(t,ts,vs,t',l,SOME F,NIL)) 
                                 = "\\ADECFO{"^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
                                              ^ prTerms vs ^ "}{"
                                              ^ prTerm t' ^ "}{"
                                              ^ prSymbol l ^ "}{"
                                              ^ prForm F ^ "}"
      | prProc (ADEC(t,ts,vs,t',l,SOME F,P)) 
                                 = "\\ADECFP{"^ prTerm t ^ "}{"
                                              ^ prTerms ts ^ "}{"
                                              ^ prTerms vs ^ "}{"
                                              ^ prTerm t' ^ "}{"
                                              ^ prSymbol l ^ "}{"
                                              ^ prForm F ^ "}\n{" 
                                              ^ prProc P ^ "}"
     
      | prProc (PAR_I(i,x,P))    = "\\PARI{" ^ i ^ "}{" 
                                              ^ int2str' x ^ "}\n{"
                                              ^ prProc P ^ "}"
      | prProc (PAR_X(i,x,j,P))  = "\\PARX{" ^ i ^ "}{" 
                                             ^ int2str' x ^ "}{" 
                                             ^ j ^ "}\n{"
                                             ^ prProc P ^ "}"
      | prProc (NEW_I((sis,t)::ts,P)) 
                                 = let val is = String.concat
						    (map (fn (s,i) => int2str' i) sis)
                                       val ss = String.concat
						    (map (fn (s,i) => s) sis)
				   in "\\NEWIS{" ^ ss ^ "}{" 
                                                ^ is ^ "}{" 
						^ prTerm t ^ "}{" 
						^ prProc (NEW_I(ts,P)) ^ "}"
				   end
      | prProc (NEW_I([],P))     = prProc P
      | prProc (ANEW_I((sis,t)::ts,P)) 
                                 = let val is = String.concat
						    (map (fn (s,i) => int2str' i) sis)
                                       val ss = String.concat
						    (map (fn (s,i) => s) sis)
				   in "\\ANEWIS{" ^ ss ^ "}{" 
                                                 ^ is ^ "}{" 
					 	 ^ prTerm t ^ "}{" 
						 ^ prProc (ANEW_I(ts,P)) ^ "}"
				   end
      | prProc (ANEW_I([],P))     = prProc P
      | prProc (LEAK(P,ls))      = "\\LEAK{" ^ prProc P ^ "}{" ^ prLs ls ^ "}"
      | prProc (TEST s)          = "\\MAGENTA{" ^ s ^ "}"

    and prProcs (P::P'::Ps)      =  prProc P ^ "\n\\\\\n \\mid & "
                                    ^ prProcs (P'::Ps)
      | prProcs [P]              =  prProc P ^ "\n"
      | prProcs []               = ""

    (* printing protocol ********)

    fun prProtocol P             = "$$\n\\begin{array}[t]{l}\n" 
				   ^ prProc P 
				   ^ "\n\\end{array}\n$$\n"


    (* printing leaks ***********)

    fun prLeak (str,(P1,P2,P3))  = str ^ ":\\\\ $\\small\\begin{array}[t]{l}" 
                                       ^ (if P1 = NIL 
					  then "" 
					  else (prProc P1 ^ "\\\\"))
                                       ^ (if P2 = NIL 
					  then "" 
					  else (prProc P2 ^ "\\\\"))
                                       ^ (if P3 = NIL 
					  then ""
					  else (prProc P3))^
				       "\\end{array}$"
				       
    fun prLeaks (at::ats)        = "\\item " ^ (prLeak at) 
                                             ^ "\n" ^ (prLeaks ats)
      | prLeaks []               = ""

    (* printing observations ****)

    fun prObs ((ls,t)::ob::obs)  = "$" ^ (prTerm t) ^ "$, " ^ (prObs (ob::obs))
      | prObs [(ls,t)]           = "$" ^ (prTerm t) ^ "$"
      | prObs []                 = ""




   (**************************************************************************)
   (* clean removes unnecessary leak information:                            *)
   (*  - true:  for protocols that have been unfolded                        *)
   (*  - false: for protocols in Meta Lysa formulation                       *) 
   (**************************************************************************) 

   fun clean b NIL                       = NIL
     | clean b (OUT(ts,P))               = OUT(ts,clean b P)
     | clean b (INP(ts,xs,P))            = INP(ts,xs,clean b P)
     | clean b (PAR Ps)                  = PAR (map (clean b) Ps) 
     | clean b (NEW(t,P))                = NEW(t,clean b P)
     | clean b (ANEW(t,P))               = ANEW(t,clean b P)
     | clean b (BANG P)                  = BANG(clean b P)
     | clean b (DEC(t,ts,xs,tk,l,Fs,P))  = DEC(t,ts,xs,tk,l,Fs,clean b P)
     | clean b (ADEC(t,ts,xs,tk,l,Fs,P)) = ADEC(t,ts,xs,tk,l,Fs,clean b P)
     | clean b (PAR_I(i,x,P))            = PAR_I(i,x,clean b P)
     | clean b (PAR_X(i,x,j,P))          = PAR_X(i,x,j,clean b P)
     | clean b (NEW_I(ts,P))             = NEW_I(ts,clean b P)
     | clean b (ANEW_I(ts,P))             = ANEW_I(ts,clean b P)
     | clean b (LEAK(TEST txt,ls))       = LEAK(TEST txt,ls)
     | clean b (LEAK(NIL,_))             = NIL
     | clean b (LEAK(P,ls))              = if b 
					   then clean b P
					   else LEAK(clean b P,ls)
     | clean b (TEST(s))                 = TEST(s)




    (***********************************************************************)
    (* Auxiliary finctions for unfolding of observation variables          *)
    (***********************************************************************)
	    


    fun unfoldObs (s,i,n) t = if i <= n 
                          then (substTerm (s,i) t) :: (unfoldObs (s,i+1,n) t)
                          else []


    fun ml2lObs' n ((s,i)::is,t) = 
	let val res = unfoldObs (s,i,n) t
        in List.concat (map (fn t' => ml2lObs' n (is,t')) res)
        end

      |  ml2lObs' n ([],t)        = [t]


    fun ml2lObs n obs = List.concat (map (ml2lObs' n) obs)




    (***********************************************************************)
    (* Auxiliary functions for transformations:                            *)
    (* substitute the base symbol in indexed names and variables           *)
    (***********************************************************************)    


    fun apply_subst ((n,n')::sub) (n'',s)
                                    = if n = n'' then (n',s)
				      else apply_subst sub (n'',s)
      | apply_subst [] (n'',s) = (n'',s)


    fun substSymbTerm sub (NAME n)        = NAME(apply_subst sub n)
      | substSymbTerm sub (NAMEP n)       = NAMEP(apply_subst sub n)
      | substSymbTerm sub (NAMEM n)       = NAMEM(apply_subst sub n)
      | substSymbTerm sub (VAR x)         = VAR (apply_subst sub x)
      | substSymbTerm sub (ENC(ts,t,l,Fs))  
	                                  = ENC(map (substSymbTerm sub) ts, 
 						substSymbTerm sub t, l, Fs)
      | substSymbTerm sub (AENC(ts,t,l,Fs))  
	                                  = AENC(map (substSymbTerm sub) ts, 
			 			 substSymbTerm sub t, l, Fs)


    fun substSymbProc sub NIL             = NIL
      | substSymbProc sub (OUT(ts,P))     = OUT(map (substSymbTerm sub) ts, 
						substSymbProc sub P)
      | substSymbProc sub (INP(ts,xs,P))  = INP(map (substSymbTerm sub) ts,
						map (substSymbTerm sub) xs,
						substSymbProc sub P)
      | substSymbProc sub (PAR Ps)        = PAR (map (substSymbProc sub) Ps)
      | substSymbProc sub (NEW(t,P))      = NEW(substSymbTerm sub t,
						substSymbProc sub P)
      | substSymbProc sub (ANEW(t,P))     = ANEW(substSymbTerm sub t,
						 substSymbProc sub P)
      | substSymbProc sub (BANG P)        = BANG(substSymbProc sub P)
      | substSymbProc sub (DEC(t,ts,xs,tk,l,Fs,P))
                                       = DEC(substSymbTerm sub t, 
					     map (substSymbTerm sub) ts, 
					     map (substSymbTerm sub) xs,
                                             substSymbTerm sub tk, l, Fs, 
                                             substSymbProc sub P)
      | substSymbProc sub (ADEC(t,ts,xs,tk,l,Fs,P))
                                       = ADEC(substSymbTerm sub t, 
					      map (substSymbTerm sub) ts, 
					      map (substSymbTerm sub) xs,
                                              substSymbTerm sub tk, l, Fs, 
                                              substSymbProc sub P)     
      | substSymbProc sub (PAR_I(i,x,P))  = PAR_I(i,x,substSymbProc sub P) 
      | substSymbProc sub (PAR_X(i,x,j,P))= PAR_X(i,x,j,substSymbProc sub P)
      | substSymbProc sub (NEW_I(ts,P))   = NEW_I(map (fn (si,t) => 
							  (si,substSymbTerm sub t)) ts,
					       substSymbProc sub P)
      | substSymbProc sub (ANEW_I(ts,P))  = ANEW_I(map (fn (si,t) => 
							   (si,substSymbTerm sub t)) ts,
						   substSymbProc sub P)
      | substSymbProc sub (LEAK(P,ls))    = LEAK(substSymbProc sub P,ls)
      | substSymbProc sub (TEST s)        = TEST s

in

(*****************************************************************************)
(* prettyPrint : Prot -> file                                                *)
(* prettyPrint prints a protocol to a file "results/lab.tex"                 *)
(*****************************************************************************)
      

fun prettyPrint (lab,name,spec,leaks,obs) = 
    let val P   = clean false (spec (NIL,NIL,NIL))
        val fil = TextIO.openOut ("results/"^lab^".tex")
        val _   = TextIO.output(fil, 
				"\\begin{table}\n\\framebox[\\textwidth]{\n" ^
				"\\begin{minipage}{\\textwidth}\n")
        val _   = TextIO.output(fil,prProtocol P)
        val _   = if leaks = []
                  then ()
                  else let val _ = TextIO.output(fil,"{\\bf Leaks:}\\begin{itemize}\n")
                           val _ = TextIO.output(fil,prLeaks leaks)
                           val _ = TextIO.output(fil,"\n\\end{itemize}\n")
                       in ()
                       end
        val _   = if obs = []
                  then ()
                  else let val _ = TextIO.output(fil,"{\\bf Observation variables:}\n")
                           val _ = TextIO.output(fil,prObs obs)
                       in ()
                       end
        val _   = TextIO.output(fil, 
                         "\\end{minipage}}\n\\caption{" ^ name ^ "}\n\\label{" ^ lab ^ "}\n" ^
                         "\\end{table}\n");        val _   = TextIO.closeOut fil
    in ()
    end


(*****************************************************************************)
(* chooseLeak: initialise the protocol to a given leak                       *)
(* if the leak number l is ~1 then a string tells where the leak will        *)
(* inserted.                                                                 *)
(*****************************************************************************)

fun chooseLeak l (lab,name,spec,leaks,obs) = 
    let val (txt,leak) = if l = 0
                         then ("No leaks",(NIL,NIL,NIL))
                         else if l = ~1
                              then ("test",(TEST "Leak A is inserted here",
					    TEST "Leak B is inserted here",
					    TEST "Leak C is inserted here"))
                              else List.nth (leaks,l-1)
    in (lab,name^": "^txt,clean false (spec leak), obs)
    end


(*****************************************************************************)
(* ml2All instantiates a MetaLysa process and observation variables          *) 
(*****************************************************************************)

fun ml2lAll n (lab,name,P,obs) = 
    (lab,name^" (unfolded)",clean true (ml2lProc n P), ml2lObs n obs)


(*****************************************************************************)
(* functions for role and key transformation                                 *)
(*****************************************************************************)

fun chooseRoleBoth (lab,name,P,obs) = 
    let val sub  = [("A","R"),("B","R"),("K^A","K"),("K^B","K")]        
        val P' = clean false (substSymbProc sub P)
    in (lab,name^" ($A=B=S$ and $K^A = K^B$)",P',obs)
    end


fun chooseRoleKey (lab,name,P,obs) = 
    let val sub  = [("K^A","K"),("K^B","K")]
	val P' = clean false (substSymbProc sub P)
    in (lab,name^" ($K^A = K^B$)",P',obs)
    end


fun chooseRoleABS (lab,name,P,obs) = 
    let val sub  = [("A","R"),("B","R")]  
        val P' = clean false (substSymbProc sub P)
    in (lab,name^" ($A=B=S$)",P',obs)
    end



fun prettyPrint2 (lab,name,spec,leaks,obs) = 
    let val P   = clean false (spec (NIL,NIL,NIL))
        val (lab,name,P,obs) = ml2lAll 2 (lab,name,P,obs)
        val (lab,name,P,obs) = chooseRoleBoth(lab,name,P,obs)
        val fil = TextIO.openOut ("results/pp2-"^lab^".tex")
        val _   = TextIO.output(fil, 
				"\\begin{table}\n\\framebox[\\textwidth]{\n" ^
				"\\begin{minipage}{\\textwidth}\n")
        val _   = TextIO.output(fil,prProtocol P)
        val _   = if leaks = []
                  then ()
                  else let val _ = TextIO.output(fil,"{\\bf Leaks:}\\begin{itemize}\n")
                           val _ = TextIO.output(fil,prLeaks leaks)
                           val _ = TextIO.output(fil,"\n\\end{itemize}\n")
                       in ()
                       end
(*        val _   = if obs = []
                  then ()
                  else let val _ = TextIO.output(fil,"{\\bf Observation variables:}\n")
                           val _ = TextIO.output(fil,prObs obs)
                       in ()
                       end*)
        val _   = TextIO.output(fil, 
                         "\\end{minipage}}\n\\caption{" ^ name ^ "}\n\\label{" ^ lab ^ "}\n" ^
                         "\\end{table}\n");        val _   = TextIO.closeOut fil
    in ()
    end

end

(*****************************************************************************) 
(*****************************************************************************) 

(* Functions below are considered to be DEPRECATED                           *)

(*****************************************************************************) 
(*****************************************************************************)



(*
    fun printTerms (fil,ts) = 
	let val _   = TextIO.output(fil, "\\section{Dolev-Yao initial knowledge}\n");
	    val _   = TextIO.output(fil,"$" ^ prTerms ts ^ "$");
	in ()
	end

    fun printObs (fil,ts) = 
	let val _   = TextIO.output(fil, "\\section{Observation variables}\n");
		val _   = TextIO.output(fil,"$" ^ prTerms ts ^ "$");
	in ()
	end



(*****************************************************************************) 
(* Main pretty printing functions                                                           
   - printProt: prints a table with the full specification of the protocol  
   - printProt': as pretty but with a slightly different layout              *)
(*****************************************************************************) 


fun printProt (fil,(lab,name,P,obs)) = 
    let val _   = TextIO.output(fil, "\\section{" ^name^"}\n");
        val _   = TextIO.output(fil,prProtocol P);
    in ()
    end



fun printProt' (fil,(lab,name,P,obs)) = 
    let fun split (NEW(t,P))       = let val (ts,b,P1,P2,P3) = split P
                                     in (t::ts,b,P1,P2,P3)
                                     end
         |  split (BANG P)         = let val (ts,b,P1,P2,P3) = split P
                                     in (ts,true,P1,P2,P3)
                                     end
         |  split (PAR [P1,P2])    = ([],false,P1,P2,NIL)
         |  split (PAR [P1,P2,P3]) = ([],false,P1,P2,P3)
        val (ts,b,P1,P2,P3) = split P
        val _   = TextIO.output(fil, "\\section{" ^name^"}\n");
        val _   = TextIO.output(fil,"$$(\\nu\\ " ^ prTerms ts ^ ")$$\\\\\n" ^
                                    (if b then "$!$" else ""));
        val _   = TextIO.output(fil,prProtocol P1);
        val _   = TextIO.output(fil,prProtocol P2);
        val _   = if P3 = NIL 
                  then ()
                  else TextIO.output(fil,prProtocol P3);
    in ()
    end

(*****************************************************************************) 
(*****************************************************************************) 
(*****************************************************************************)
(* Unfolding protocol to distinguish between old and current run             *)
(*   - substSymbTerm performs a substitution of names and variables in terms    *)
(*   - old_symb construct old versions of symbols                            *)
(*   - old_Proc renames all bound names and variables to be old              *)
(*   - makeOld takes a protocol and makes an old version running in parallel *)
(*     with a new version                                                    *)
(*****************************************************************************) 





fun old_symb (NAME(n,ss)::xs) = (n,"o"^n)::(old_symb xs)
 |  old_symb (VAR(n,ss)::xs)  = (n,"o"^n)::(old_symb xs)
 |  old_symb [] = []

fun old_Proc theta NIL             = NIL
 |  old_Proc theta (OUT(ts,P))     = OUT(map (substSymbTerm theta) ts, 
                                         old_Proc theta P)
 |  old_Proc theta (INP(ts,xs,P))  = let val theta' = old_symb xs
                                     in INP(map (substSymbTerm theta) ts, 
                                            map (substSymbTerm theta') xs,
                                            old_Proc (theta'@theta) P)
                                     end
 |  old_Proc theta (PAR Ps)        = PAR (map (old_Proc theta) Ps)
 |  old_Proc theta (NEW(t,P))      = let val theta' = old_symb [t]
                                     in NEW(substSymbTerm theta' t,
                                            old_Proc (theta'@theta) P)
                                     end
 |  old_Proc theta (BANG P)        = BANG(old_Proc theta P)
 |  old_Proc theta (DEC(t,ts,xs,tk,l,Fs,P))
                                   = let val theta' = old_symb xs
                                     in DEC(substSymbTerm theta t, 
                                            map (substSymbTerm theta) ts, 
                                            map (substSymbTerm theta') xs,
                                            substSymbTerm theta tk, l, Fs, 
                                            old_Proc (theta'@theta) P)
                                     end
 |  old_Proc theta (PAR_I(s,i,P))  = PAR_I(s,i,old_Proc theta P) 
 |  old_Proc theta (PAR_X(s,i,s',P)) 
                                   = PAR_X(s,i,s',old_Proc theta P)
 |  old_Proc theta (LEAK(P,ls))    = LEAK(old_Proc theta P,ls)
 |  old_Proc theta (TEST s)        = TEST s

fun old_Obs ((ls,VAR(n,ss))::xs)   = (ls,VAR(n,ss))::
                                     (ls,VAR("o"^n,ss))::
                                     (old_Obs xs)
 |  old_Obs []                     = []

(*****************************************************************************) 
(* main functions for creating rounds                                        *)
(*  - oldversion initialises the protocol to a given leak and introduced an  *)
(*    old round with the leaks                                               *)
(*****************************************************************************) 

fun oldversion v (lab,name,spec,leaks,obs) = 
    let val (txt,leak) = if v = 0
                         then ("no leaks",(NIL,NIL,NIL))
                         
                         else if v = ~1
                              then ("test",(TEST "A",TEST "B",TEST "S"))
                              else List.nth (leaks,v-1)
        val NEW_I(ts,BANG P)  = spec (NIL,NIL,NIL)
        val NEW_I(_,BANG P')  = spec leak
        val prot  = clean false 
                      (NEW_I(ts,PAR[BANG(old_Proc [] P'),P]))
    in (lab,name^" with rounds: "^txt,prot, old_Obs obs)
    end


*)



(*

fun normal n v spec =
    let val fil    = TextIO.openOut "out.txt";
        
	val prot1  = version v spec
        val _      = printProt(fil,prot1)
        val prot2  = unfold n prot1
        val _      = printProt'(fil,prot2)
        val dy     = dolev_yao prot2
        val _      = printTerms (fil,dy)
        val (_,_,_,obs) = prot2
        val _      = printObs (fil,ml2lObs n obs)
    in TextIO.closeOut fil
    end

fun rounds n v spec =
    let val fil    = TextIO.openOut "out.txt";
        val prot1  = oldversion v spec
        val _      = printProt(fil,prot1)
        val prot2  = unfold n prot1
        val _      = printProt'(fil,prot2)
        val dy     = dolev_yao prot2
        val _      = printTerms (fil,dy)
        val (_,_,_,obs) = prot2
        val _      = printObs (fil,ml2lObs n obs)
    in TextIO.closeOut fil
    end

fun noroles n v spec =
    let val fil    = TextIO.openOut "out.txt";
        val prot1  = roles_both (version v spec)
        val _      = printProt(fil,prot1)
        val prot2  = unfold n prot1
        val _      = printProt'(fil,prot2)
        val dy     = dolev_yao prot2
        val _      = printTerms (fil,dy)
        val (_,_,_,obs) = prot2
        val _      = printObs (fil,ml2lObs n obs)
    in TextIO.closeOut fil
    end

fun noABSroles n v spec =
    let val fil    = TextIO.openOut "out.txt";
        val prot1  = roles_ABS (version v spec)
        val _      = printProt(fil,prot1)
        val prot2  = unfold n prot1
        val _      = printProt'(fil,prot2)
        val dy     = dolev_yao prot2
        val _      = printTerms (fil,dy)
        val (_,_,_,obs) = prot2
        val _      = printObs (fil,ml2lObs n obs)
    in TextIO.closeOut fil
    end

fun noKEYroles n v spec =
    let val fil    = TextIO.openOut "out.txt";
        val prot1  = roles_key (version v spec)
        val _      = printProt(fil,prot1)
        val prot2  = unfold n prot1
        val _      = printProt'(fil,prot2)
        val dy     = dolev_yao prot2
        val _      = printTerms (fil,dy)
        val (_,_,_,obs) = prot2
        val _      = printObs (fil,ml2lObs n obs)
    in TextIO.closeOut fil
    end


fun pretty (lab,name,spec,leaks,obs) = 
    let val P   = clean false (spec (NIL,NIL,NIL))
                  (* (TEST "A",TEST "B",TEST "S") *)
        val fil = TextIO.openOut "results/out.txt";
        val _   = TextIO.output(fil, 
                         "\\begin{table}\n\\framebox[\\textwidth]{\n" ^
                         "\\begin{minipage}{\\textwidth}\n" ^
                         "{\\bf Protocol:}\n");
        val _   = TextIO.output(fil,prProtocol P);
        val _   = if leaks = []
                  then ()
                  else let val _ = TextIO.output(fil,"{\\bf Leaks:}\\begin{itemize}\n")
                           val _ = TextIO.output(fil,prLeaks leaks)
                           val _ = TextIO.output(fil,"\n\\end{itemize}\n")
                       in ()
                       end
        val _   = if obs = []
                  then ()
                  else let val _ = TextIO.output(fil,"{\\bf Observation variables:}\n")
                           val _ = TextIO.output(fil,prObs obs)
                       in ()
                       end
        val _   = TextIO.output(fil, 
                         "\\end{minipage}}\n\\caption{" ^ name ^ "}\n\\label{" ^ lab ^ "}\n" ^
                         "\\end{table}\n");
        val _   = TextIO.closeOut fil
    in ()
    end

*)

end
