(*****************************************************************************) 
(* Andrew Secure RPC handshake protocol                                      *)
(*****************************************************************************) 
structure Andrew =
struct
local
    open Lysa
    open Protocols
in
 
val NA  = NAME ("N^A",["i","j"])
val NB  = NAME ("N^B",["i","j"])
val K'  = NAME ("K'",["i","j"])
val NB' = NAME ("N^{B'}",["i","j"])

(* variables **************************)

val y   = VAR  ("y",  ["i","j"])
val yNb = VAR  ("y^{N_B}",  ["i","j"])
val u   = VAR  ("u",  ["i","j"])
val uK  = VAR  ("u^K",  ["i","j"])
val uN  = VAR  ("u^N",  ["i","j"])
val x   = VAR  ("x",  ["i","j"])
val xNa = VAR  ("x^{N_A}",  ["i","j"])
val z   = VAR  ("z",  ["i","j"])
val v   = VAR  ("v",  ["i","j"])
val vm  = VAR  ("v^m",["i","j"])

val oK  = NAME ("oK_{12}",[])
val oN  = NAME ("oN_{12}",[])
val oNA = NAME ("oN^A_{12}",[])
val om  = NAME ("om_{12}",[])


(*****************************************************************************) 

val a1 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1),("j",1)],K)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(NA,
                           OUT([PA,A,PB,B,PA,A,ENC([NA],K,LA1,dest LB1)], 
                           INP([PB,B,PA,A],[y],
                           DEC(y,[ENC([NA],succ,LB,dest LA)],[yNb],K,LA2,orig LB2,
                           OUT([PA,A,PB,B,ENC([ENC([yNb],succ,LA,dest LB)],K,LA3,dest LB3)],
                           INP([PB,B,PA,A],[u],
                           DEC(u,[],[uK,uN],K,LA4,orig LB4,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],uK,LA,dest LB)], 
                           LEAK(leakA,[("i",1)]))))))))))))         
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",0,
                           INP([PA,A,PB,B,PA,A],[x], 
                           DEC(x,[],[xNa],K,LB1,orig LA1,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([ENC([xNa],succ,LB,dest LA),NB],K,LB2,dest LA2)],
                           INP([PA,A,PB,B],[z],
                           DEC(z,[ENC([NB],succ,LA,dest LB)],[],K,LB3,orig LA3,
                           NEW(K',
                           NEW(NB',
                           OUT([PB,B,PA,A,ENC([K',NB'],K,LB4,dest LA4)],
                           INP([PA,A,PB,B],[v],
                           DEC(v,[zero],[vm],K',LB,orig LA,
                           LEAK(leakB,[("i",1),("j",2)]))))))))))))))
                val Swmf = LEAK(leakS,[])
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("a1","Andrew Secure RPC handshake",spec,
        [("Session key $K_{12}$ leaked",
         (NIL,
          OUT([PB,B,P0,B,NAME("K",["i","j"])],NIL),
          NIL)),
         ("old session key $oK_{12}$",  
               (NIL,
                NEW(oK,NEW(oN,
                           OUT([PS,S,P0,B,oK,oN,
                                ENC([oK,oN],K,LB,dest LA)],
                               NIL))),
                NIL))],
        [([("i",1),("j",1)],vm)]
        ) 
    end 

(*****************************************************************************) 

val a2 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1),("j",1)],K)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(NA,
                           OUT([PA,A,PB,B,PA,A,ENC([one,NA],K,LA1,dest LB1)], 
                           INP([PB,B,PA,A],[y],
                           DEC(y,[two,ENC([NA],succ,LB,dest LA)],[yNb],K,LA2,orig LB2,
                           OUT([PA,A,PB,B,ENC([three,ENC([yNb],succ,LA,dest LB)],K,LA3,dest LB3)],
                           INP([PB,B,PA,A],[u],
                           DEC(u,[four],[uK,uN],K,LA4,orig LB4,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],uK,LA,dest LB)], 
                           LEAK(leakA,[("i",1)]))))))))))))                         
                
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",0,
                           INP([PA,A,PB,B,PA,A],[x], 
                           DEC(x,[one],[xNa],K,LB1,orig LA1,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([two,ENC([xNa],succ,LB,dest LA),NB],K,LB2,dest LA2)],
                           INP([PA,A,PB,B],[z],
                           DEC(z,[three,ENC([NB],succ,LA,dest LB)],[],K,LB3,orig LA3,
                           NEW(K',
                           NEW(NB',
                           OUT([PB,B,PA,A,ENC([four,K',NB'],K,LB4,dest LA4)],
                           INP([PA,A,PB,B],[v],
                           DEC(v,[zero],[vm],K',LB,orig LA,
                           LEAK(leakB,[("i",1),("j",2)]))))))))))))))
                val Swmf = LEAK(leakS,[])
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("a2","Andrew Secure RPC handshake (type flaw fixed)",spec,
        [("Session key $K_{12}$ leaked",
         (NIL,
          OUT([PB,B,P0,B,NAME("K",["i","j"])],NIL),
          NIL)),
         ("old session key $oK_{12}$",  
               (NIL,
                NEW(oK,NEW(oN,
                           OUT([PS,S,P0,B,oK,oN,
                                ENC([four,oK,oN],K,LB4,dest LA4)],
                               NIL))),
                NIL))],
        [([("i",1),("j",1)],vm)]
        ) 
    end 

(*****************************************************************************) 

val a3 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1),("j",1)],K)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(NA,
                           OUT([PA,A,PB,B,PA,A,ENC([one,NA],K,LA1,dest LB1)], 
                           INP([PB,B,PA,A],[y],
                           DEC(y,[two,ENC([NA],succ,LB,dest LA)],[yNb],K,LA2,orig LB2,
                           OUT([PA,A,PB,B,ENC([three,ENC([yNb],succ,LA,dest LB)],K,LA3,dest LB3)],
                           INP([PB,B,PA,A],[u],
                           DEC(u,[four,NA],[uK,uN],K,LA4,orig LB4,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],uK,LA,dest LB)], 
                           LEAK(leakA,[("i",1)]))))))))))))         
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",0,
                           INP([PA,A,PB,B,PA,A],[x], 
                           DEC(x,[one],[xNa],K,LB1,orig LA1,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([two,ENC([xNa],succ,LB,dest LA),NB],K,LB2,dest LA2)],
                           INP([PA,A,PB,B],[z],
                           DEC(z,[three,ENC([NB],succ,LA,dest LB)],[],K,LB3,orig LA3,
                           NEW(K',
                           NEW(NB',
                           OUT([PB,B,PA,A,ENC([four,xNa,K',NB'],K,LB4,dest LA4)],
                           INP([PA,A,PB,B],[v],
                           DEC(v,[zero],[vm],K',LB,orig LA,
                           LEAK(leakB,[("i",1),("j",2)]))))))))))))))
                val Swmf = LEAK(leakS,[])
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("a3","Andrew Secure RPC handshake (BAN correction)",spec,
        [("Session key $K_{12}$ leaked",
         (NIL,
          OUT([PB,B,P0,B,NAME("K",["i","j"])],NIL),
          NIL)),
         ("old session key $oK_{12}$",  
               (NIL,
                NEW(oK,NEW(oN,NEW(oNA,
                           OUT([PS,S,P0,B,oK,oN,oNA,
                                ENC([four,oNA,oK,oN],K,LB4,dest LA4)],
                               NIL)))),
                NIL))],
        [([("i",1),("j",1)],vm)]
        ) 
    end 
end
end