(*****************************************************************************)
(* MSR protocol (Modular Square Root)                                        *)
(*                                                                           *)
(* M. J. Beller, Li-Fung Chang, and  Y. Yacobi: Privacy and Authentication   *)
(* on a Protable Communications System. IEEE Journal of Selected Areas in    *)
(* Communications. 11(6). p. 821-829. 1993.                                  *)
(*                                                                           *)
(* MSR, and IMSR protocols -- for wits                                       *)
(*                                               26.11.2003 Mikael Buchholtz *)
(*****************************************************************************)

structure MSRW =
struct
local
    open Lysa
    open Protocols
in

val KUP   = NAMEP("KU",[])
val KUM   = NAMEM("KU",[])


val K    = NAME("K",["i","j"])

val xKB  = VAR("xKB",["i","j"])
val xCS  = VAR("xCS",["i","j"])
val xC   = VAR("xC", ["i","j"])





val R     = NAME("R",[])
val CR    = NAME("CR",["j"])

val KRP   = NAMEP("KB",["j"])
val KRM   = NAMEM("KB",["j"])

val yEK   = VAR("y1",["i","j"])
val yK    = VAR("y2",["i","j"])
val yEC   = VAR("y3",["i","j"])
val yC    = VAR("y4",["i","j"])


val LA1   = ("a1",["i"])
val LA2   = ("a2",["i"])
val LA3   = ("a3",["i"])
val LA4   = ("a4",["i"])


val LB1   = ("b1",["j"])
val LB2   = ("b2",["j"])
val LB3   = ("b3",["j"])
val LB4   = ("b4",["j"])

val A    = NAME("A",["i"])
val B    = NAME("B",["j"])

val LUA   = ("ua",["i"])
val LUB   = ("ub",["j"])



fun porig l = NONE (* SOME (ORIG [l,CPDY]) *)
fun pdest l = NONE (* SOME (DEST [l,CPDY]) *)


(*****************************************************************************)
val p1 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,
			    PAR_I("j",1,
			    INP([B,A,B],[xKB],
			    NEW(K,
                            OUT([A,B,AENC([K],xKB,LA2,dest LB2)],
			    OUT([A,B,ENC([A,AENC([A],KUM,LA3, pdest LB4)],
					  K,LA4,dest LB3)],
				NIL))))))

                val BProc = ANEW_I([([("j",1)], NAME("KB",["j"]))],
		            BANG(
		            PAR_I("j",1,
			    PAR_I("i",1,
                            OUT([B,A,B,KRP],
			    INP([A,B],[yEK],
                            ADEC(yEK,[],[yK],KRM,LB2,porig(LA1),
                            INP([A,B],[yEC],
			    DEC(yEC,[A],[yC],yK,LB3,orig LA4,
			    ADEC(yC,[A],[],KUP,LB4,orig LA3,
		            NIL))))))))))

                val SProc = OUT([KUP,KUP,KUP],NIL)

             in	ANEW(NAME("KU",[]),		
		PAR [BANG(AProc),BProc,SProc])
             end
    in ("msrw1",	
	"MSR with $n$ portables $A_i$ and $n$ base stations $B_{j}$",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yK)	 
	 ])
	   
    end 
(*****************************************************************************)



(*****************************************************************************)
val p2 =
    let fun spec (leakA : Lysa.Proc,leakB : Lysa.Proc,leakS : Lysa.Proc) =
	    let val AProc = PAR_I("i",1,
			    PAR_I("j",1,
			    BANG(
			    INP([B,A,B],[xC,xKB],
			    ADEC(xC,[B],[],KUP,LA1,orig LB1,
			    NEW(K,
                            OUT([A,B,AENC([K],xKB,LA2,dest LB2)],
			    OUT([A,B,ENC([A,AENC([A],KUM,LA3, pdest LB4)],
					  K,LA4,dest LB3)],
				NIL))))))))

                val BProc = ANEW_I([([("j",1)], NAME("KB",["j"]))],		            
		            PAR_I("j",1,
			    PAR_I("i",1,
			    BANG(
                            OUT([B,A,B,AENC([B],KUM,LB1,pdest LA1),KRP],
			    INP([A,B],[yEK],
                            ADEC(yEK,[],[yK],KRM,LB2,porig(LA2),
                            INP([A,B],[yEC],
			    DEC(yEC,[A],[yC],yK,LB3,orig LA4,
			    ADEC(yC,[A],[],KUP,LB4,orig LA3,
		            NIL))))))))))

                val SProc = OUT([KUP,KUP,KUP],NIL)

             in	ANEW(NAME("KU",[]),		
		PAR [AProc,BProc,SProc])
             end
    in ("msrw2",
	"Improved MSR with $n$ portables $A_i$ and $n$ base stations $B_{j}$",
	spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yK)	 
	 ])
	   
    end 
(*****************************************************************************)

end
end(*;

pl MSRW.p1;*)