(*****************************************************************************) 
(* Needham Schroeder symmetric key protocol                                  *)
(*  - has type flaw problem and leaked old session key problem               *)
(*****************************************************************************) 
structure NS =
struct

local
    open Lysa
    open Protocols
in

val NA  = NAME ("N^A",["i","j"])
val NB  = NAME ("N^B",["i","j"])
val x   = VAR  ("x",  ["i","j"])
val y   = VAR  ("y",  ["i","j"])
val yK  = VAR  ("y^K",["i","j"])
val yM  = VAR  ("y'", ["i","j"])
val z   = VAR  ("z",  ["i","j"])
val zK  = VAR  ("z^K",["i","j"])
val u   = VAR  ("u",  ["i","j"])
val uN  = VAR  ("u^N",["i","j"])
val v   = VAR  ("v",  ["i","j"])
val vN  = VAR  ("v^N",["i","j"])
val msg = NAME ("m",["i","j"])
val w   = VAR  ("w",  ["i","j"])
val wM  = VAR  ("w^m",["i","j"])

val oK   = NAME ("oK_{12}",  [])
val oN   = NAME ("oN_{12}",  [])
val oNA  = NAME ("oN^A_{12}",[])
val oNB  = NAME ("oN^B_{12}",[])
val om   = NAME ("om_{12}",  [])

val ns1 = let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(NA,
                           OUT([PA,A,PS,S,PA,A,PB,B,NA], 
                           INP([PS,S,PA,A],[y], 
                           DEC(y,[PB,B,NA],[yK,yM],KA,LA2,orig LS2,
                           OUT([PA,A,PB,B,yM], 
                           INP([PB,B,PA,A],[u], 
                           DEC(u,[],[uN],yK,LA4,orig LB4,
                           OUT([PA,A,PB,B,
                                ENC([ENC([uN],succ,LAs,dest LBs)],yK,LA5,dest LB5)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],yK,LA6,dest LB6)], 
                           LEAK(leakA,[("i",1)])))))))))))))
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PA,A,PB,B],[z], 
                           DEC(z,[PA,A],[zK],KB,LB3,orig LS3,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([NB],zK,LB4,dest LA4)], 
                           INP([PA,A,PB,B],[v], 
                           DEC(v,[ENC([NB],succ,LAs,dest LBs)],[],zK,LB5,orig LA5,
                           INP([PA,A,PB,B],[w], 
                           DEC(w,[zero],[wM],zK,LB6,orig LA6,
                           LEAK(leakB,[("j",1)])))))))))))
                val Swmf = PAR_I("i",0,
                           PAR_I("j",0,
                           INP([PA,A,PS,S,PA,A,PB,B],[x], 
                           NEW(K,
                           OUT([PS,S,PA,A,
                                ENC([PB,B,x,K,ENC([PA,A,K],KB,LS3,dest LB3)],
                                KA,LS2,dest LA2)], 
                           LEAK(leakS,[("i",1),("j",2)])))))) 
               in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
        in ("ns1","Needham Schroeder symmetric key protocol",spec,
            [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
             ("master key $K^A_1$",  (OUT([PB,B,P0,A,KA],NIL), NIL, NIL)),
             ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
             ("old session key, A- and B-nonces and all encryptions",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,NEW(oNB,NEW(om,
                   OUT([PS,S,P0,A,oK,oNA,oNB],
                   OUT([PS,S,P0,A,
                        ENC([PB,B,oNA,oK,ENC([PA,A,oK],KB,LS3,dest LB3)],
                                         KA,LS2,dest LA2)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oK],KB,LS3,dest LB3)],
                   OUT([PS,S,P0,A,
                        ENC([oNB],oK,LB4,dest LA4),
                        ENC([ENC([oNB],succ,LAs,dest LBs)],oK,LA5,dest LB5)],
                   OUT([PS,S,P0,A,
                        ENC([zero,om],oK,LA6,dest LB6)],
                 NIL))))))))))),
             ("old session key and two tickets",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,
                  OUT([PS,S,P0,A,oK],
                  OUT([PS,S,P0,A,
                       ENC([PB,B,oNA,oK,ENC([PA,A,oK],KB,LS3,dest LB3)],
                                         KA,LS2,dest LA2)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oK],KB,LS3,dest LB3)],
                 NIL)))))))],
               [([("i",1),("j",1)],wM)])
           end 

(*****************************************************************************) 
(* Needham Schroeder symmetric key protocol                                  *)
(*  - type flaw problem corrected but still leaked old session key problem   *)    
(*****************************************************************************) 

val ns3 = let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(NA,
                           OUT([PA,A,PS,S,PA,A,PB,B,NA], 
                           INP([PS,S,PA,A],[y], 
                           DEC(y,[PB,B,NA],[yK,yM],KA,LA2,orig LS2,
                           OUT([PA,A,PB,B,yM], 
                           INP([PB,B,PA,A],[u], 
                           DEC(u,[one],[uN],yK,LA4,orig LB4,
                           OUT([PA,A,PB,B,
                                ENC([two,ENC([uN],succ,LAs,dest LBs)],yK,LA5,dest LB5)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],yK,LA6,dest LB6)], 
                           LEAK(leakA,[("i",1)])))))))))))))
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PA,A,PB,B],[z], 
                           DEC(z,[PA,A],[zK],KB,LB3,orig LS3,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([one,NB],zK,LB4,dest LA4)], 
                           INP([PA,A,PB,B],[v], 
                           DEC(v,[two,ENC([NB],succ,LAs,dest LBs)],[],zK,LB5,orig LA5,
                           INP([PA,A,PB,B],[w], 
                           DEC(w,[zero],[wM],zK,LB6,orig LA6,
                           LEAK(leakB,[("j",1)])))))))))))
                val Swmf = PAR_I("i",0,
                           PAR_I("j",0,
                           INP([PA,A,PS,S,PA,A,PB,B],[x], 
                           NEW(K,
                           OUT([PS,S,PA,A,
                                ENC([PB,B,x,K,ENC([PA,A,K],KB,LS3,dest LB3)],
                                KA,LS2,dest LA2)], 
                           LEAK(leakS,[("i",1),("j",2)])))))) 
               in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
        in ("ns3","Needham Schroeder (type flaw corrected)",spec,
            [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
             ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
             ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
             ("old session key, A- and B-nonces and all encryptions",
                (NIL,NIL,                 NEW(oK,NEW(oNA,NEW(oNB,NEW(om,
                   OUT([PS,S,P0,A,oK,oNA,oNB],
                   OUT([PS,S,P0,A,
                        ENC([PB,B,oNA,oK,ENC([PA,A,oK],KB,LS3,dest LB3)],
                                         KA,LS2,dest LA2)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oK],KB,LS3,dest LB3)],
                   OUT([PS,S,P0,A,
                        ENC([one,oNB],oK,LB4,dest LA4),
                        ENC([two,ENC([oNB],succ,LAs,dest LBs)],oK,LA5,dest LB5)],
                   OUT([PS,S,P0,A,
                        ENC([zero,om],oK,LA6,dest LB6)],
                 NIL))))))))))),
             ("old session key and two tickets",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,
                  OUT([PS,S,P0,A,oK],
                  OUT([PS,S,P0,A,
                       ENC([PB,B,oNA,oK,ENC([PA,A,oK],KB,LS3,dest LB3)],
                                         KA,LS2,dest LA2)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oK],KB,LS3,dest LB3)],
                 NIL)))))))],
               [([("i",1),("j",1)],wM)])
           end 



(*****************************************************************************) 
(*****************************************************************************) 
(*****************************************************************************) 
(*****************************************************************************) 
(*****************************************************************************) 
(*****************************************************************************) 
(* Amended Needham Schroeder symmetric key protocol                          *)
(*  - has type flaw problem; old session key problem corrected               *)
(*****************************************************************************) 

val t   = VAR ("t",["i","j"])
val N   = NAME ("N",["i","j"])
val x'  = VAR ("x'",["i","j"])
val xN  = VAR ("x^N",["i","j"])

val oNB' = NAME ("oN'_{12}",[])

val ns2 = let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           OUT([PA,A,PB,B,PA,A],
                           INP([PB,B,PA,A],[t],
                           NEW(NA,
                           OUT([PA,A,PS,S,PA,A,PB,B,NA,t], 
                           INP([PS,S,PA,A],[y], 
                           DEC(y,[PB,B,NA],[yK,yM],KA,LA4,orig LS4,
                           OUT([PA,A,PB,B,yM], 
                           INP([PB,B,PA,A],[u], 
                           DEC(u,[],[uN],yK,LA6,orig LB6,
                           OUT([PA,A,PB,B,ENC([ENC([uN],succ,LAs,dest LBs)],
                                yK,LA7,dest LB7)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],yK,LA8,dest LB8)], 
                           LEAK(leakA,[("i",1)])))))))))))))))
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PA,A,PB,B,PA,A],[],
                           NEW(N,
                           OUT([PB,B,PA,A,ENC([PA,A,N],KB,LB2,dest LS2)],
                           INP([PA,A,PB,B],[z], 
                           DEC(z,[PA,A,N],[zK],KB,LB5,orig LS5,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([NB],zK,LB6,dest LA6)], 
                           INP([PA,A,PB,B],[v], 
                           DEC(v,[ENC([NB],succ,LAs,dest LBs)],[],zK,LB7,orig LA7,
                           INP([PA,A,PB,B],[w], 
                           DEC(w,[zero],[wM],zK,LB8,orig LA8,
                           LEAK(leakB,[("j",1)]))))))))))))))
                val Swmf = PAR_I("i",0,
                           PAR_I("j",0,
                           INP([PA,A,PS,S,PA,A,PB,B],[x,x'], 
                           DEC(x',[PA,A],[xN],KB,LS2,orig LB2,
                           NEW(K,
                           OUT([PS,S,PA,A,
                                ENC([PB,B,x,K,ENC([PA,A,xN,K],KB,LS5,dest LB5)],
                                KA,LS4,dest LA4)], 
                           LEAK(leakS,[("i",1),("j",2)])))))) )
               in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
        in ("ns2","Amended Needham Schroeder protocol",spec,
            [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
             ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
             ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
             ("old session key, A- and two B-nonces and all encryptions",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,NEW(oNB,NEW(oNB',NEW(om,
                   OUT([PS,S,P0,A,
                        oK,oNA,oNB,oNB'],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oNB'],KB,LB2,dest LA2)],
                   OUT([PS,S,P0,A,
                        ENC([PB,B,oNA,oK,ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                                         KA,LS4,dest LA4)],                   OUT([PS,S,P0,A,
                        ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                   OUT([PS,S,P0,A,
                        ENC([oNB],oK,LB6,dest LA6),
                        ENC([ENC([oNB],succ,LAs,dest LBs)],oK,LA7,dest LB7)],
                   OUT([PS,S,P0,A,
                        ENC([zero,om],oK,LA8,dest LB8)],
                 NIL))))))))))))),
             ("old session key and two tickets",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,NEW(oNB',OUT([PS,S,P0,A,
                        oK],
                   OUT([PS,S,P0,A,
                        ENC([PB,B,oNA,oK,ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                                         KA,LS4,dest LA4)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                 NIL))))))))],
               [([("i",1),("j",1)],wM)])
           end 


(*****************************************************************************) 
(* Amended Needham Schroeder symmetric key protocol                          *)
(*  - type flaw corrected                                                    *)
(*****************************************************************************) 

val ns4 = let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           OUT([PA,A,PB,B,PA,A],
                           INP([PB,B,PA,A],[t],
                           NEW(NA,
                           OUT([PA,A,PS,S,PA,A,PB,B,NA,t], 
                           INP([PS,S,PA,A],[y], 
                           DEC(y,[PB,B,NA],[yK,yM],KA,LA4,orig LS4,
                           OUT([PA,A,PB,B,yM], 
                           INP([PB,B,PA,A],[u], 
                           DEC(u,[one],[uN],yK,LA6,orig LB6,
                           OUT([PA,A,PB,B,ENC([two,ENC([uN],succ,LAs,dest LBs)],
                                yK,LA7,dest LB7)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],yK,LA8,dest LB8)], 
                           LEAK(leakA,[("i",1)])))))))))))))))
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PA,A,PB,B,PA,A],[],
                           NEW(N,
                           OUT([PB,B,PA,A,ENC([PA,A,N],KB,LB2,dest LS2)],
                           INP([PA,A,PB,B],[z], 
                           DEC(z,[PA,A,N],[zK],KB,LB5,orig LS5,
                           NEW(NB,
                           OUT([PB,B,PA,A,ENC([one,NB],zK,LB6,dest LA6)], 
                           INP([PA,A,PB,B],[v], 
                           DEC(v,[two,ENC([NB],succ,LAs,dest LBs)],[],zK,LB7,orig LA7,
                           INP([PA,A,PB,B],[w], 
                           DEC(w,[zero],[wM],zK,LB8,orig LA8,
                           LEAK(leakB,[("j",1)]))))))))))))))
                val Swmf = PAR_I("i",0,
                           PAR_I("j",0,                           INP([PA,A,PS,S,PA,A,PB,B],[x,x'], 
                           DEC(x',[PA,A],[xN],KB,LS2,orig LB2,
                           NEW(K,
                           OUT([PS,S,PA,A,
                                ENC([PB,B,x,K,ENC([PA,A,xN,K],KB,LS5,dest LB5)],
                                KA,LS4,dest LA4)], 
                           LEAK(leakS,[("i",1),("j",2)])))))) )
               in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
        in ("ns4","Amended Needham Schroeder (with type flaw corrected)",spec,
            [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
             ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
             ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
             ("old session key, A- and two B-nonces and all encryptions",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,NEW(oNB,NEW(oNB',NEW(om,
                   OUT([PS,S,P0,A,
                        oK,oNA,oNB,oNB'],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oNB'],KB,LB2,dest LA2)],
                   OUT([PS,S,P0,A,
                        ENC([PB,B,oNA,oK,ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                                         KA,LS4,dest LA4)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                   OUT([PS,S,P0,A,
                        ENC([one,oNB],oK,LB6,dest LA6),
                        ENC([two,ENC([oNB],succ,LAs,dest LBs)],oK,LA7,dest LB7)],
                   OUT([PS,S,P0,A,
                        ENC([zero,om],oK,LA8,dest LB8)],
                 NIL))))))))))))),
             ("old session key and two tickets",
                (NIL,NIL,
                 NEW(oK,NEW(oNA,NEW(oNB',OUT([PS,S,P0,A,
                        oK],
                   OUT([PS,S,P0,A,
                        ENC([PB,B,oNA,oK,ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                                         KA,LS4,dest LA4)],
                   OUT([PS,S,P0,A,
                        ENC([PA,A,oNB',oK],KB,LS5,dest LB5)],
                 NIL))))))))],
               [([("i",1),("j",1)],wM)])
           end 

end
end
(*****************************************************************************) 

