(*****************************************************************************)

structure NSPK =
struct
local
    open Lysa
    open Protocols
in

val zPB  = VAR("zI",["j"])
val zB   = VAR("zB",["j"])
val xEKB = VAR("x1",["i"])
val xKB  = VAR("xKB",["i","j"])
val xENANB  = VAR("x2",["i","j"])
val xNB  = VAR("xNB",["i","j"])

val yEA  = VAR("y1",["j"])
val yEKA = VAR("y2",["j"])
val yKA  = VAR("yKA",["j","i"])
val yNA  = VAR("yNA",["j","i"])
val yENB = VAR("y3",["j","i"])
val yEEM = VAR("y4",["j","i"])
val yEM  = VAR("y5",["j","i"])
val yM   = VAR("y^m",["j","i"])


val KAP = NAMEP("K^A",["i"])
val KAM = NAMEM("K^A",["i"])
val KBP = NAMEP("K^B",["j"])
val KBM = NAMEM("K^B",["j"])

val KSP = NAMEP("KS",[])
val KSM = NAMEM("KS",[])

val NA = NAME("nA",["i","j"])
val NB = NAME("nB",["j","i"])

val LS1   = ("S1",[])
val LS2   = ("S2",[])

(*
val LA1   = ("A1",["i","j"])
val LA2   = ("A2",["i","j"])
val LA3   = ("A3",["i","j"])
val LA4   = ("A4",["i","j"])
val LA5   = ("A5",["i","j"])
val LA6   = ("A6",["i","j"])


val LB1   = ("B1",["i","j"])
val LB2   = ("B2",["i","j"])
val LB3   = ("B3",["i","j"])
val LB4   = ("B4",["i","j"])
val LB5   = ("B5",["i","j"])
val LB6   = ("B6",["i","j"])
*)

val LA1   = ("A1",["i"])
val LA2   = ("A2",["i"])
val LA3   = ("A3",["i"])
val LA4   = ("A4",["i"])
val LA5   = ("A5",["i"])
val LA6   = ("A6",["i"])


val LB1   = ("B1",["j"])
val LB2   = ("B2",["j"])
val LB3   = ("B3",["j"])
val LB4   = ("B4",["j"])
val LB5   = ("B5",["j"])
val LB6   = ("B6",["j"])


fun porig l = SOME (ORIG [l,CPDY]);

(*****************************************************************************)
val p1 =
    let fun spec (leakA,leakB,leakS) =
	    let val AProc = PAR_I("i",1,
			    PAR_X("j",0,"i",
			    OUT([PA,A,PS,S,PA,A,PB,B],				
			    INP([PS,S,PA,A],[xEKB],			    
			    ADEC(xEKB,[PB,B],[xKB],KSP,("A1",[]),orig LS1,
		            NEW(NA,
			    OUT([PA,A,PB,B,AENC([PA,A,NA],xKB,LA2,dest LB1)],
                            INP([PB,B,PA,A],[xENANB],
                            ADEC(xENANB,[NA],[xNB],KAM,LA3,porig LB3,
			    OUT([PA,A,PB,B,AENC([xNB],xKB,LA4,dest LB4)],
			    
                            LEAK(leakA,[("i",1)])))))))))))

                val BProc = PAR_I("j",1,
			    PAR_X("i",0,"i", 
		            INP([PA,A,PB,B],[yEA],
			    ADEC(yEA,[PA,A],[yNA],KBM,LB1,porig LA2,
                            OUT([PB,B,PS,S,PB,B,PA,A],
			    INP([PS,S,PB,B],[yEKA],
			    ADEC(yEKA,[PA,A],[yKA],KSP,("B2",[]),orig LS2,
		            NEW(NB,
			    OUT([PB,B,PA,A,AENC([yNA,NB],yKA,LB3,dest LA3)],
			    INP([PA,A,PB,B],[yENB],
                            ADEC(yENB,[NB],[],KBM,LB4,porig LA4,
                            
			    LEAK(leakB,[("j",1)]))))))))))))

                val SProc = PAR[
			    OUT([KSP],
			    OUT([NAMEM("K^A",["0"])],
			    OUT([NAMEM("K^B",["0"])],NIL))),
			    PAR_I("i",0,
		            PAR_I("j",0,
			    INP([PA,A,PS,S,PA,A,PB,B],[],
		            OUT([PS,S,PA,A,
				 AENC([PB,B,KBP],KSM,LS1,
				      SOME(DEST([("A1",[]),CPDY])))],
		            LEAK(leakS,[("i",1),("j",2)]))))),
			    PAR_I("j",0,
		            PAR_I("i",0,
			    INP([PB,B,PS,S,PB,B,PA,A],[],
		            OUT([PS,S,PB,B,
				 AENC([PA,A,KAP],KSM,LS2,
				      SOME(DEST([("B2",[]),CPDY])))],
		            NIL))))]

             in ANEW_I([([("i",1)],NAME("K^A",["i"]))],
		ANEW_I([([("j",1)],NAME("K^B",["j"]))],
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("nspk1","Needham-Schroeder Public Key",spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yNA),	 
	 ([("i",1),("j",1)],xNB)
	 ])
	   
    end 
(*****************************************************************************)


val p2 =
    let fun spec (leakA,leakB,leakS) =
	    let val AProc = PAR_I("i",1,
			    PAR_X("j",0,"i",			    
		            NEW(NA,
			    OUT([PA,A,PB,B,AENC([PA,A,NA],KBP,LA2,dest LB1)],
                            INP([PB,B,PA,A],[xENANB],
                            ADEC(xENANB,[NA],[xNB],KAM,LA3,porig LB3,
			    OUT([PA,A,PB,B,AENC([xNB],KBP,LA4,dest LB4)],		    

                            LEAK(leakA,[("i",1)]))))))))

                val BProc = PAR_I("j",1,
			    PAR_X("i",0,"i", 
		            INP([PA,A,PB,B],[yEA],
			    ADEC(yEA,[PA,A],[yNA],KBM,LB1,porig LA2,
                            NEW(NB,
			    OUT([PB,B,PA,A,AENC([yNA,NB],KAP,LB3,dest LA3)],
			    INP([PA,A,PB,B],[yENB],
                            ADEC(yENB,[NB],[],KBM,LB4,porig LA4,

			    LEAK(leakB,[("j",1)])))))))))

                val SProc = PAR_I("i",0,
			    PAR_I("j",0,
                            OUT([KAP,
				 KBP,
				 NAMEM("K^A",["0"]),
				 NAMEM("K^B",["0"]),
				 NAME("I",["0"])],
		            LEAK(leakS,[("i",1),("j",2)]))))

             in ANEW_I([([("i",1)],NAME("K^A",["i"]))],
		ANEW_I([([("j",1)],NAME("K^B",["j"]))],
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("nspk2","Needham-Schroeder Public Key (no Server)",spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yNA),	 
	 ([("i",1),("j",1)],xNB)
	 ])
	   
    end
(*****************************************************************************)


(*****************************************************************************)
val p3 =
    let fun spec (leakA,leakB,leakS) =
	    let val AProc = PAR_I("i",1,
			    PAR_X("j",0,"i",
			    OUT([PA,A,PS,S,PA,A,PB,B],				
			    INP([PS,S,PA,A],[xEKB],			    
			    ADEC(xEKB,[PB,B],[xKB],KSP,("A1",[]),orig LS1,
		            NEW(NA,
			    OUT([PA,A,PB,B,AENC([PA,A,NA],xKB,LA2,dest LB1)],
                            INP([PB,B,PA,A],[xENANB],
                            ADEC(xENANB,[PB,B,NA],[xNB],KAM,LA3,porig LB3,
			    OUT([PA,A,PB,B,AENC([xNB],xKB,LA4,dest LB4)],
			    
                            LEAK(leakA,[("i",1)])))))))))))

                val BProc = PAR_I("j",1,
			    PAR_X("i",0,"i", 
		            INP([PA,A,PB,B],[yEA],
			    ADEC(yEA,[PA,A],[yNA],KBM,LB1,porig LA2,
                            OUT([PB,B,PS,S,PB,B,PA,A],
			    INP([PS,S,PB,B],[yEKA],
			    ADEC(yEKA,[PA,A],[yKA],KSP,("B2",[]),orig LS2,
		            NEW(NB,
			    OUT([PB,B,PA,A,AENC([PB,B,yNA,NB],yKA,LB3,dest LA3)],
			    INP([PA,A,PB,B],[yENB],
                            ADEC(yENB,[NB],[],KBM,LB4,porig LA4,
                            
			    LEAK(leakB,[("j",1)]))))))))))))

                val SProc = PAR[
			    OUT([KSP],
			    OUT([NAMEM("K^A",["0"])],
			    OUT([NAMEM("K^B",["0"])],NIL))),
			    PAR_I("i",0,
		            PAR_I("j",0,
			    INP([PA,A,PS,S,PA,A,PB,B],[],
		            OUT([PS,S,PA,A,
				 AENC([PB,B,KBP],KSM,LS1,
				      SOME(DEST([("A1",[]),CPDY])))],
		            LEAK(leakS,[("i",1),("j",2)]))))),
			    PAR_I("j",0,
		            PAR_I("i",0,
			    INP([PB,B,PS,S,PB,B,PA,A],[],
		            OUT([PS,S,PB,B,
				 AENC([PA,A,KAP],KSM,LS2,
				      SOME(DEST([("B2",[]),CPDY])))],
		            NIL))))]

             in ANEW_I([([("i",1)],NAME("K^A",["i"]))],
		ANEW_I([([("j",1)],NAME("K^B",["j"]))],
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("nspk3","Lowe's amended Needham-Schroeder Public Key",spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yNA),	 
	 ([("i",1),("j",1)],xNB)
	 ])
	   
    end 
(*****************************************************************************)

val p4 =
    let fun spec (leakA,leakB,leakS) =
	    let val AProc = PAR_I("i",1,
			    PAR_X("j",0,"i",			    
		            NEW(NA,
			    OUT([PA,A,PB,B,AENC([PA,A,NA],KBP,LA2,dest LB1)],
                            INP([PB,B,PA,A],[xENANB],
                            ADEC(xENANB,[PB,B,NA],[xNB],KAM,LA3,porig LB3,
			    OUT([PA,A,PB,B,AENC([xNB],KBP,LA4,dest LB4)],		    

                            LEAK(leakA,[("i",1)]))))))))

                val BProc = PAR_I("j",1,
			    PAR_X("i",0,"i", 
		            INP([PA,A,PB,B],[yEA],
			    ADEC(yEA,[PA,A],[yNA],KBM,LB1,porig LA2,
                            NEW(NB,
			    OUT([PB,B,PA,A,AENC([PB,B,yNA,NB],KAP,LB3,dest LA3)],
			    INP([PA,A,PB,B],[yENB],
                            ADEC(yENB,[NB],[],KBM,LB4,porig LA4,

			    LEAK(leakB,[("j",1)])))))))))

                val SProc = PAR_I("i",0,
			    PAR_I("j",0,
                            OUT([KAP,
				 KBP,
				 NAMEM("K^A",["0"]),
				 NAMEM("K^B",["0"]),
				 NAME("I",["0"])],
		            LEAK(leakS,[("i",1),("j",2)]))))

             in ANEW_I([([("i",1)],NAME("K^A",["i"]))],
		ANEW_I([([("j",1)],NAME("K^B",["j"]))],
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("nspk4","Lowe's amended Needham-Schroeder Public Key (no Server)",spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yNA),	 
	 ([("i",1),("j",1)],xNB)
	 ])
	   
    end
(*****************************************************************************)




val p6 =
    let fun spec (leakA,leakB,leakS) =
	    let val AProc = PAR_I("i",1,
			    PAR_X("j",0,"i",			    
		            NEW(NA,
			    OUT([PA,A,PB,B,AENC([PA,A,NA],KBP,LA2,dest LB1)],
                            INP([PB,B,PA,A],[xENANB],
                            ADEC(xENANB,[NA],[xNB],KAM,LA3,NONE,
			    OUT([PA,A,PB,B,AENC([xNB],KBP,LA4,dest LB4)],		    

                            LEAK(leakA,[("i",1)]))))))))

                val BProc = PAR_I("j",1,
			    PAR_X("i",0,"i", 
		            INP([PA,A,PB,B],[yEA],
			    ADEC(yEA,[PA,A],[yNA],KBM,LB1,NONE,
                            NEW(NB,
			    OUT([PB,B,PA,A,AENC([yNA,NB],KAP,LB3,dest LA3)],
			    INP([PA,A,PB,B],[yENB],
                            ADEC(yENB,[NB],[],KBM,LB4,NONE,

			    LEAK(leakB,[("j",1)])))))))))

                val SProc = PAR_I("i",0,
			    PAR_I("j",0,
                            OUT([KAP,
				 KBP,
				 NAMEM("K^A",["0"]),
				 NAMEM("K^B",["0"]),
				 NAME("I",["0"])],
		            LEAK(leakS,[("i",1),("j",2)]))))

             in ANEW_I([([("i",1)],NAME("K^A",["i"]))],
		ANEW_I([([("j",1)],NAME("K^B",["j"]))],
		ANEW(NAME("KS",[]),
		PAR [BANG(AProc),BANG(BProc),BANG(SProc)])))
             end
    in ("nspk6","Needham-Schroeder Public Key (no Server) -- experiments with annotations",spec,
	[("",(NIL,NIL,NIL))],
	[([("i",1),("j",1)],yNA),	 
	 ([("i",1),("j",1)],xNB)
	 ])
	   
    end
(*****************************************************************************)

end
end