(*****************************************************************************) 
(* Otway-Rees protocol                                                       *)
(*****************************************************************************) 
structure Otway =
struct
local
    open Lysa
    open Protocols
in

val N   = NAME ("N"  ,["i","j"])
val NA  = NAME ("N^A",["i","j"])
val NB  = NAME ("N^B",["i","j"])

val x    = VAR ("x^A",["i","j"])
val xN   = VAR ("x^N",["i","j"])

val yN   = VAR ("y^N",    ["j"])
val yA   = VAR ("y^A",    ["j"])
val yB   = VAR ("y^B",    ["j"])
val yAn  = VAR ("y^{A_N}",["i","j"])
val yBn  = VAR ("y^{B_N}",["i","j"])
val zA   = VAR ("z^A",    ["i","j"])
val zB   = VAR ("z^B",    ["i","j"])
val zK   = VAR ("z^K",    ["i","j"])
val v    = VAR ("v",      ["i","j"])
val vK   = VAR ("v^K",    ["i","j"])
val w    = VAR ("w",      ["i","j"])
val wM   = VAR ("w^m",    ["i","j"])

val oK   = NAME ("oK_{12}",  [])
val oN   = NAME ("oN_{12}",  [])
val oNA  = NAME ("oN^A_{12}",[])
val oNB  = NAME ("oN^B_{12}",[])

val om   = NAME ("om_{12}", [])

val or1  = let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(N,
                           NEW(NA,
                           OUT([PA,A,PB,B,N,
                                ENC([PA,A,PB,B,N,NA],KA,LA1,dest LS1)], 
                           INP([PB,B,PA,A,N],[v], 
                           DEC(v,[NA],[vK],KA,LA3,orig LS3,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],vK,LA5,dest LB5)], 
                           LEAK(leakA,[("i",1)]))))))))))
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PA,A,PB,B],[xN,x], 
                           NEW(NB,
                           OUT([PB,B,PS,S,xN,x,
                                ENC([PA,A,PB,B,xN,NB],KB,LB2,dest LS2)],
                           INP([PS,S,PB,B,xN],[zA,zB], 
                           DEC(zB,[NB],[zK],KB,LB4,orig LS4,
                           OUT([PB,B,PA,A,xN,zA],
                           INP([PA,A,PB,B],[w], 
                           DEC(w,[zero],[wM],zK,LB5,orig LA5,
                           LEAK(leakB,[("j",1)])))))))))))
                val Swmf = PAR_I("j",0,
                           INP([PB,B,PS,S],[yN,yA,yB], 
                           PAR_I("i",0,
                           DEC(yA,[PA,A,PB,B,yN],[yAn],KA,LS1,orig LA1,
                           DEC(yB,[PA,A,PB,B,yN],[yBn],KB,LS2,orig LB2,
                           NEW(K,
                           OUT([PS,S,PB,B,yN,ENC([yAn,K],KA,LS3,dest LA3),
                                  ENC([yBn,K],KB,LS4,dest LB4)], 
                           LEAK(leakS,[("i",1),("j",2)]))))))))
            in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
    in ("or1","Otway Rees",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
         ("old session key, all nonces and encryptions",
              (NIL,NIL,
               NEW(oK,NEW(oNA,NEW(oNB,NEW(oN,NEW(om,
                OUT([PS,S,P0,B,oN,oNA,oNB,oK],
                OUT([PS,S,P0,B,           
                     ENC([PA,A,PB,B,oN,oNA],KA,LA1,dest LS1)],
                OUT([PS,S,P0,B,           
                     ENC([PA,A,PB,B,oN,oNB],KB,LB2,dest LS2)],
                OUT([PS,S,P0,B,           
                     ENC([oNA,oK],KA,LS3,dest LA3)],
                OUT([PS,S,P0,B,           
                     ENC([oNB,oK],KB,LS4,dest LB4)],
                OUT([PS,S,P0,B,           
                     ENC([zero,om],oK,LA5,dest LB5)],
                   NIL))))))))))))),
         ("old session key and tickets",
              (NIL,NIL,
               NEW(oK,NEW(oNA,NEW(oNB,
                OUT([PS,S,P0,B,oK],
                OUT([PS,S,P0,B,           
                     ENC([oNA,oK],KA,LS3,dest LA3)],                OUT([PS,S,P0,B,           
                     ENC([oNB,oK],KB,LS4,dest LB4)],
                   NIL))))))))],
        (*[([("i",1),("j",1)],wM)])*)
	[([("i",1),("j",1)],yAn),([("i",1),("j",1)],yBn)])
    end 

(*****************************************************************************)

val yBn'  = VAR ("y^{B_N}",["j"])

val or4  = let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(N,
                           OUT([PA,A,PB,B,N,
                                ENC([PA,A,PB,B,N],KA,LA1,dest LS1)], 
                           INP([PB,B,PA,A,N],[v], 
                           DEC(v,[N],[vK],KA,LA4,orig LS4,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],vK,LA5,dest LB5)], 
                           LEAK(leakA,[("i",1)])))))))))
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PA,A,PB,B],[xN,x], 
                           NEW(NB,
                           OUT([PB,B,PS,S,xN,x,NB,
                                ENC([PA,A,PB,B,xN],KB,LB2,dest LS2)],
                           INP([PS,S,PB,B,xN],[zA,zB], 
                           DEC(zB,[NB],[zK],KB,LB3,orig LS3,
                           OUT([PB,B,PA,A,xN,zA],
                           INP([PA,A,PB,B],[w], 
                           DEC(w,[zero],[wM],zK,LB5,orig LA5,
                           LEAK(leakB,[("j",1)])))))))))))
                val Swmf = PAR_I("j",0,
                           INP([PB,B,PS,S],[yN,yA,yBn',yB], 
                           PAR_I("i",0,
                           DEC(yA,[PA,A,PB,B,yN],[],KA,LS1,orig LA1,
                           DEC(yB,[PA,A,PB,B,yN],[],KB,LS2,orig LB2,
                           NEW(K,
                           OUT([PS,S,PB,B,yN,ENC([yN,K],KA,LS4,dest LA4),
                                  ENC([yBn',K],KB,LS3,dest LB3)], 
                           LEAK(leakS,[("i",1),("j",2)]))))))))
            in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
    in ("or4","Otway Rees (BAN optimised)",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
         ("old session key $oK_{12}$",      (NIL,NIL,
                                       NEW(oK,NEW(oNB,NEW(oN,
                                         OUT([PS,S,P0,B,oN,oK,oNB],
                                         OUT([PS,S,P0,B,
                                            ENC([PA,A,PB,B,oN],KA,LA1,dest LS1)],
                                         OUT([PS,S,P0,B,
                                            ENC([PA,A,PB,B,oN],KB,LB2,dest LS2)],
                                         OUT([PS,S,P0,B,
                                            ENC([oN,oK],KA,LS4,dest LA4)],
                                         OUT([PS,S,P0,B,                                            ENC([oNB,oK],KB,LS3,dest LB3)],
                                               NIL))))))))))],
        [([("i",1),("j",1)],wM)])
    end 

(*****************************************************************************)
(* Otway Rees with hashing                                                   *)
(*  - exhibits a weakness of the analysis of matching                        *)
(*****************************************************************************)

val vAk  = VAR ("v^{A_K}",["i","j"])
val vA   = VAR ("v^{A}" ,["i","j"])
val zAk  = VAR ("z^{A_K}",["i","j"])
val zA   = VAR ("z^{A}" ,["i","j"])
val zBk  = VAR ("z^{B_K}",["i","j"])
val zB   = VAR ("z^{B}" ,["i","j"])
val xAn  = VAR ("x^{A_N}",["i","j"])
val yN   = VAR ("y^N",    ["i","j"])
val yA   = VAR ("y^A",    ["i","j"])
val yB   = VAR ("y^B",    ["i","j"])

val or2 = let fun spec (leakA,leakB,leakS) =
           let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
               val Awmf = PAR_I ("i",1,
                          PAR_X ("j",1,"i",
                          NEW(N,
                          NEW(NA,
                          OUT([PA,A,PB,B,N,NA,
                               ENC([PA,A,PB,B,N,NA],KA,LA1,dest LS1)], 
                          INP([PB,B,PA,A,N],[vAk,vA], 
                          DEC(vA,[NA,vAk],[],KA,LA,orig LS,
                          DEC(vAk,[],[vK],KA,LA,orig LS,
                          NEW(msg,
                          OUT([PA,A,PB,B,ENC([zero,msg],vK,LA,dest LB)], 
                          LEAK(leakA,[("i",1)])))))))))))
               val Bwmf = PAR_I("j",1,
                          PAR_I("i",~1,
                          INP([PA,A,PB,B],[xN,xAn,x], 
                          NEW(NB,
                          OUT([PB,B,PS,S,PA,A,xN,xAn,NB,x,
                               ENC([PA,A,PB,B,xN,NB],KB,LB2,dest LS2)],
                          INP([PS,S,PB,B,xN],[zAk,zA,zBk,zB], 
                          DEC(zB,[NB,zBk],[],KB,LB,orig LS,
                          DEC(zBk,[],[zK],KB,LB,orig LS,
                          OUT([PB,B,PA,A,xN,zAk,zA],
                          INP([PA,A,PB,B],[w], 
                          DEC(w,[zero],[wM],zK,LB,orig LA,
                          LEAK(leakB,[("j",1)]))))))))))))
               val Swmf = PAR_I("i",0,
                          PAR_I("j",0,
                          INP([PB,B,PS,S,PA,A],[yN,yAn,yBn,yA,yB], 
                          DEC(yA,[PA,A,PB,B,yN,yAn],[],KA,LS1,orig LA1,
                          DEC(yB,[PA,A,PB,B,yN,yBn],[],KB,LS2,orig LB2,
                          NEW(K,
                          OUT([PS,S,PB,B,yN,
                              ENC([K],KA,LS3,dest LA3),
                              ENC([yAn,ENC([K],KA,LS3,dest LA3)],KA,LS4,dest LA4),
                              ENC([K],KB,LS5,dest LB5),
                              ENC([yBn,ENC([K],KB,LS5,dest LB5)],KB,LS6,dest LB6)],
                          LEAK(leakS,[("i",1),("j",2)]))))))))
            in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
           in ("or2","Otway-Rees protocol - with hashing",spec,
               [(*("session key leaked by S", 
                    (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
                ("master key $K^A_1$ leaked",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
                ("master key $K^B_1$ leaked",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
                ("old session key $oK_{12}$ leaked",(NIL,NIL,
                    NEW(oK,NEW(oNA,NEW(oNB,NEW(oN,
                        OUT([PS,S,P0,B,oK,oN,oNA,oNB],
                        OUT([PS,S,P0,B,
                              ENC([PA,A,PB,B,oN,oNA],KA,LA1,dest LS1)],
                        OUT([PS,S,P0,B,
                              ENC([PA,A,PB,B,oN,oNB],KB,LB2,dest LS2)],
                        OUT([PS,S,P0,B,
                              ENC([oK],KA,LS3,dest LA3)],
                        OUT([PS,S,P0,B,
                              ENC([oNA,ENC([oK],KA,LS3,dest LA3)],KA,LS4,dest LA4)],
                        OUT([PS,S,P0,B,
                              ENC([oK],KB,LS5,dest LB5)],
                        OUT([PS,S,P0,B,
                              ENC([oNB,ENC([oK],KB,LS5,dest LB5)],KB,LS6,dest LB6)],
                        NIL)))))))))))))],
                ([([("i",1),("j",1)],wM)]))
           end 

(*****************************************************************************)
(* Otway Rees with hashing                                                   *)
(*  - a "repair" of the protocol to circumvent the analysis weakness         *)
(*****************************************************************************)

val vAk'  = VAR ("v^{A_K:A}",["i","j"])
val zBk'  = VAR ("z^{B_K:B}",["i","j"])

val yN'   = VAR ("y^{N:S}",    ["i","j"])
val yN''  = VAR ("y^{N:SS}",    ["i","j"])
val yAn'  = VAR ("y^{A_N:S}",["i","j"])
val yBn'  = VAR ("y^{B_N:S}",["i","j"])

val or3 = let fun spec (leakA,leakB,leakS) =
           let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
               val Awmf = PAR_I ("i",1,
                          PAR_X ("j",1,"i",
                          NEW(N,
                          NEW(NA,
                          OUT([PA,A,PB,B,N,NA,
                               ENC([PA,A,PB,B,N,NA],KA,LA,dest LS)], 
                          INP([PB,B,PA,A,N],[vAk,vA], 
                          DEC(vA,[NA,vAk],[],KA,LA,orig LS,
                          DEC(vA,[NA],[vAk'],KA,LA,orig LS,
                          DEC(vAk',[],[vK],KA,LA,orig LS,
                          NEW(msg,
                          OUT([PA,A,PB,B,ENC([msg],vK,LA,dest LB)], 
                          LEAK(leakA,[("i",1)]))))))))))))
               val Bwmf = PAR_I("j",1,
                          PAR_I("i",~1,
                          INP([PA,A,PB,B],[xN,xAn,x], 
                          NEW(NB,
                          OUT([PB,B,PS,S,PA,A,xN,xAn,NB,x,
                               ENC([PA,A,PB,B,xN,NB],KB,LB,dest LS)],
                          INP([PS,S,PB,B,xN],[zAk,zA,zBk,zB], 
                          DEC(zB,[NB,zBk],[],KB,LB,orig LS,
                          DEC(zB,[NB],[zBk'],KB,LB,orig LS,
                          DEC(zBk',[],[zK],KB,LB,orig LS,
                          OUT([PB,B,PA,A,xN,zAk,zA],
                          INP([PA,A,PB,B],[w], 
                          DEC(w,[],[wM],zK,LB,orig LA,
                          LEAK(leakB,[("j",1)])))))))))))))
               val Swmf = PAR_I("i",0,
                          PAR_I("j",0,
                          INP([PB,B,PS,S,PA,A],[yN,yAn,yBn,yA,yB], 
                          DEC(yA,[PA,A,PB,B,yN,yAn],[],KA,LS,orig LA,
                          DEC(yA,[PA,A,PB,B],[yN',yAn'],KA,LS,orig LA,
                          DEC(yB,[PA,A,PB,B,yN',yBn],[],KB,LS,orig LB,
                          DEC(yB,[PA,A,PB,B],[yN'',yBn'],KB,LS,orig LB,
                          NEW(K,
                          OUT([PS,S,PB,B,yN'',
                              ENC([K],KA,LS,dest LA),
                              ENC([yAn',ENC([K],KA,LS,dest LA)],KA,LS,dest LA),
                              ENC([K],KB,LS,dest LB),
                              ENC([yBn',ENC([K],KB,LS,dest LB)],KB,LS,dest LB)],
                          LEAK(leakS,[("i",1),("j",2)]))))))))))
            in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
            end
           in ("or3","Otway-Rees protocol - with hashing and extra match",spec,
               [(*("session key leaked by S", 
                    (NIL,NIL,OUT([PS,S,P0,B,K],NIL))),*)
                ("master key $K^A_1$ leaked",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
                ("master key $K^B_1$ leaked",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
                ("old session key $oK_{12}$ leaked",(NIL,NIL,
                    NEW(oK,NEW(oNA,NEW(oNB,NEW(oN,
                        OUT([PS,S,P0,B,oK,oN,oNA,oNB,
                              ENC([PA,A,PB,B,oN,oNA],KA,LA,dest LS),
                              ENC([PA,A,PB,B,oN,oNB],KB,LB,dest LS),
                              ENC([oK],KA,LS,dest LA),
                              ENC([oNA,ENC([oK],KA,LS,dest LA)],KA,LS,dest LA),
                              ENC([oK],KB,LS,dest LB),
                              ENC([oNB,ENC([oK],KB,LS,dest LB)],KB,LS,dest LB)],
                        NIL)))))))],
                ([([("i",1),("j",1)],wM)]))
           end 
end
end
(*****************************************************************************) 

