(********************************************************************)
(* Auxilliary definitions for specification of protocols.

   Partially adapted from implementation by Hanne Riis Nielson.
                                       03.09.2003. Mikael Buchholtz *)
(********************************************************************)


structure Protocols =
struct
open Lysa


fun prettyPrintList ls = List.map AuxEx1.prettyPrint ls


(* dest and orig **********************)

fun dest l = SOME (DEST [l])
fun orig l = SOME (ORIG [l]);


(* roles ******************************)

val A   = NAME ("A", [])
val B   = NAME ("B", [])
val S   = NAME ("S", [])

val R   = NAME ("R", [])

(* principals *************************)

val PA  = NAME ("I", ["i"])
val PB  = NAME ("I", ["j"])
val PS  = NAME ("I", ["-1"])
val P0  = NAME ("I", ["0"])

(* master keys ************************)

val KA  = NAME ("K^A", ["i"])
val KB  = NAME ("K^B", ["j"])

(* session keys and messages **********)

val K   = NAME ("K",   ["i","j"])
 
(* messages ***************************)

val msg = NAME ("m",  ["i","j"])

(* crypto-points **********************)

val LA  = ("A", ["i"])
val LB  = ("B", ["j"])
val LS  = ("S", [])

val LA1   = ("A^1",   ["i"])
val LA2   = ("A^2",   ["i"])
val LA3   = ("A^3",   ["i"])
val LA4   = ("A^4",   ["i"])
val LA5   = ("A^5",   ["i"])
val LA6   = ("A^6",   ["i"])
val LA7   = ("A^7",   ["i"])
val LA8   = ("A^8",   ["i"])
val LAp   = ("A^p",   ["i"])
val LAs   = ("A^s",   ["i"])
val LB1   = ("B^1",   ["j"])
val LB2   = ("B^2",   ["j"])
val LB3   = ("B^3",   ["j"])
val LB4   = ("B^4",   ["j"])
val LB5   = ("B^5",   ["j"])
val LB6   = ("B^6",   ["j"])
val LB7   = ("B^7",   ["j"])
val LB8   = ("B^8",   ["j"])
val LBp   = ("B^p",   ["j"])
val LBs   = ("B^s",   ["j"])

val LS1   = ("S^1",   [])
val LS2   = ("S^2",   [])
val LS3   = ("S^3",   [])
val LS3'  = ("S^{3'}",[])
val LS4   = ("S^4",   [])
val LS5   = ("S^5",   [])
val LS6   = ("S^6",   [])

(* numbers ****************************)

val succ  = NAME ("succ",[])
val zero  = NAME ("0",[])
val one   = NAME ("u_1",[])
val two   = NAME ("u_2",[])
val three = NAME ("u_3",[])
val four  = NAME ("u_4",[])

(* data structures ********************)

val pair = NAME ("pair",[])

end