
structure WMF =
struct
local
    open Lysa
    open Protocols
in

(*use "transform.sml";*)

(*****************************************************************************) 
(* Wide Mouth Frog protocol                                                  *)
(*  - wmf1:  no time stamps (vulnerable when old keys are leaked)            *)
(*  - wmf2:  receivers name not encrypted (flawed)                           *)
(*  - wmf3:  repair of wmf1 using nonces                                     *)
(*****************************************************************************) 
 
(* variables **************************)

val x   = VAR  ("x",  ["i"])
val xK  = VAR  ("x^K",["i","j"])
val y   = VAR  ("y",  ["j"])
val yK  = VAR  ("y^K",["i","j"])
val z   = VAR  ("z",  ["i","j"])
val zm  = VAR  ("z^m",["i","j"])

val oK  = NAME ("oK_{12}",[])
val oN  = NAME ("oN_{12}",[])
val om  = NAME ("om_{12}",[])


(*****************************************************************************) 
(* no time stamps                                                            *)
(* - vulnerable to leaked old session keys                                   *)
(*****************************************************************************) 

val wmf1 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(K,
                           OUT([PA,A,PS,S,PA,A,ENC([PB,B,K],KA,LA1,dest LS1)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],K,LA3,dest LB3)], 
                           LEAK(leakA,[("i",1)])))))))            
                val Bwmf = PAR_I("j",1,
                           INP([PS,S,PB,B],[y], 
                           PAR_I("i",~1,
                           DEC(y,[PA,A],[yK],KB,LB2,orig LS2,
                           INP([PA,A,PB,B],[z],
                           DEC(z,[zero],[zm],yK,LB3,orig LA3,
                           LEAK(leakB,[("j",1)])))))))
                val Swmf = PAR_I("i",0,
                           INP([PA,A,PS,S,PA,A],[x], 
                           PAR_I("j",0,
                           DEC(x,[PB,B],[xK],KA,LS1,orig LA1,
                           OUT([PS,S,PB,B,ENC([PA,A,xK],KB,LS2,dest LB2)],
                           LEAK(leakS,[("i",1),("j",2)]))))))
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("wmf1","Wide Mouthed Frog",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),*)
         ("old session key and tickets",  
               (NIL,NIL,
                 NEW(oK,
                   OUT([PS,S,P0,B,oK,
                        ENC([PB,B,oK],KA,LA1,dest LS1),
                        ENC([PA,A,oK],KB,LS2,dest LB2)],
                   NIL))))],
        [([("i",1),("j",1)],zm)])
    end 

(*****************************************************************************) 
(* responder not encrypted                                                   *)
(* - introduces a flaw                                                       *)
(*****************************************************************************) 

val x   = VAR  ("x",  ["i","j"])

val wmf2 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(K,
                           OUT([PA,A,PS,S,PA,A,PB,B,ENC([K],KA,LA1,dest LS1)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],K,LA3,dest LB3)], 
                           LEAK(leakA,[("i",1)])))))))            
                val Bwmf = PAR_I("j",1,
                           INP([PS,S,PB,B],[y], 
                           PAR_I("i",~1,
                           DEC(y,[PA,A],[yK],KB,LB2,orig LS2,
                           INP([PA,A,PB,B],[z],
                           DEC(z,[zero],[zm],yK,LB3,orig LA3,
                           LEAK(leakB,[("j",1)])))))))
                val Swmf = PAR_I("i",0,
                           PAR_I("j",0,
                           INP([PA,A,PS,S,PA,A,PB,B],[x], 
                           DEC(x,[],[xK],KA,LS1,orig LA1,
                           OUT([PS,S,PB,B,ENC([PA,A,xK],KB,LS2,dest LB2)],
                           LEAK(leakS,[("i",1),("j",2)]))))))
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("wmf2","Wide Mouthed Frog (with responder flaw)",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
         ("old session key and tickets",      
              (NIL,NIL,
                   NEW(oK,
                       OUT([PS,S,P0,B,oK,
                            ENC([K],KA,LA1,dest LS1),
                            ENC([PA,A,oK],KB,LS2,dest LB2)],
                        NIL)))) *)] : (string * (Proc * Proc * Proc)) list,
        [([("i",1),("j",1)],zm)])
    end 
 
(*****************************************************************************) 
(* initiator not encrypted                                                   *)
(* - introduces a flaw                                                       *)
(*****************************************************************************) 

val x   = VAR  ("x",  ["i"])
val y   = VAR  ("y",  ["i","j"])

val wmf3 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(K,
                           OUT([PA,A,PS,S,PA,A,ENC([PB,B,K],KA,LA1,dest LS1)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],K,LA3,dest LB3)], 
                           LEAK(leakA,[("i",1)])))))))            
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PS,S,PB,B,PA,A],[y], 
                           DEC(y,[],[yK],KB,LB2,orig LS2,
                           INP([PA,A,PB,B],[z],
                           DEC(z,[zero],[zm],yK,LB3,orig LA3,
                           LEAK(leakB,[("j",1)])))))))
                val Swmf = PAR_I("i",0,
                           INP([PA,A,PS,S,PA,A],[x], 
                           PAR_I("j",0,
                           DEC(x,[PB,B],[xK],KA,LS1,orig LA1,
                           OUT([PS,S,PB,B,PA,A,ENC([xK],KB,LS2,dest LB2)],
                           LEAK(leakS,[("i",1),("j",2)]))))))
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("wmf3","Wide Mouthed Frog (with initiator flaw)",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
         ("old session key and tickets",      
              (NIL,NIL,
                   NEW(oK,
                       OUT([PS,S,P0,B,oK,
                            ENC([PB,B,K],KA,LA1,dest LS1),
                            ENC([oK],KB,LS2,dest LB2)],
                       NIL)))) *)] : (string * (Proc * Proc * Proc)) list,
        [([("i",1),("j",1)],zm)])
    end 
 
(*****************************************************************************) 
(* with nonces                                                               *)
(* - repair of the leaked old session key problem                            *)
(* - the analysis gives pessimistic result since it is independent attribute *) 
(*****************************************************************************) 

val N   = NAME ("N",  ["i","j"])
val xN  = VAR  ("x^N",["i","j"])
val y   = VAR  ("y",  ["j"])
val yK  = VAR  ("y^K",["i","j"])
val yN  = VAR  ("y^N",["i","j"])
val z   = VAR  ("z",  ["i","j"])
val zK  = VAR  ("z^K",["i","j"])
val u   = VAR  ("u",  ["i","j"])
val um  = VAR  ("u^m",["i","j"])

val wmf4 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(N,
                           OUT([PA,A,PB,B,PA,A,N],
                           INP([PS,S,PA,A],[z],
                           DEC(z,[PB,B,N],[zK],KA,LA2,orig LS2,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],zK,LA3,dest LB3)],
                           LEAK(leakA,[("i",1)])))))))))          
               val Bwmf = PAR_I ("j",1,
                          PAR_I ("i",~1,
                          INP([PA,A,PB,B,PA,A],[xN],
                          NEW(K,
                          OUT([PB,B,PS,S,ENC([PA,A,xN,K],KB,LB1,dest LS1)],
                          INP([PA,A,PB,B],[u],
                          DEC(u,[zero],[um],K,LB3,orig LA3,
                          LEAK(leakB,[("j",1)]))))))))
               val Swmf = PAR_I ("j",0,
                          INP([PB,B,PS,S],[y],
                          PAR_I ("i",0,
                          DEC(y,[PA,A],[yN,yK],KB,LS1,orig LB1,
                          OUT([PS,S,PA,A,ENC([PB,B,yN,yK],KA,LS2,dest LA2)],
                          LEAK(leakS,[("i",1),("j",2)]))))))
         in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
         end
    in ("wmf4","Wide Mouthed Frog (with nonces)",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),*)
         ("old session key and tickets",      (NIL,NIL,
                                        NEW(oK,
                                        NEW(oN,
                                         OUT([PS,S,P0,B,oK, 
                                          ENC([PA,A,oN,oK],KB,LB2,dest LS2),
                                          ENC([PB,B,oN,oK],KA,LS2,dest LA2)],
                                             NIL))))) ],
        [([("i",1),("j",1)],um)])
    end

(*****************************************************************************) 
(* with nonces and pair                                                      *)
(* - repair of the leaked old session key problem                            *)
(* - the analysis gives correct result                                       *) 
(*****************************************************************************) 

val y' = VAR ("y'",["i","j"])
val z' = VAR ("z'",["i","j"])

val wmf5 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(N,
                           OUT([PA,A,PB,B,PA,A,N],
                           INP([PS,S,PA,A],[z],
                           DEC(z,[PB,B],[z'],KA,LA,orig LS,
                           DEC(z',[N],[zK],pair,LA,orig LB,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],zK,LA,dest LB)],
                           LEAK(leakA,[("i",1)]))))))))))          
               val Bwmf = PAR_I ("j",1,
                          PAR_I ("i",~1,
                          INP([PA,A,PB,B,PA,A],[xN],
                          NEW(K,
                          OUT([PB,B,PS,S,
                                ENC([PA,A,ENC([xN,K],pair,LB,dest LA)],
                                    KB,LB,dest LS)],
                          INP([PA,A,PB,B],[u],
                          DEC(u,[zero],[um],K,LB,orig LA,
                          LEAK(leakB,[("j",1)]))))))))
               val Swmf = PAR_I ("j",0,
                          INP([PB,B,PS,S],[y],
                          PAR_I ("i",0,
                          DEC(y,[PA,A],[y'],KB,LS,orig LB,
                          OUT([PS,S,PA,A,ENC([PB,B,y'],KA,LS,dest LA)],
                          LEAK(leakS,[("i",1),("j",2)]))))))
         in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))         end
    in ("wmf5","Wide Mouthed Frog (with nonces and pair)",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),*)
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
         ("old session key $oK_{12}$",      (NIL,NIL,
                                        NEW(oK,
                                        NEW(oN,
                                         OUT([PS,S,P0,B,oK,oN,
                                          ENC([PA,A,ENC([oN,oK],pair,LB,dest LA)],
                                               KB,LB,dest LS),
                                          ENC([PB,B,ENC([oN,oK],pair,LB,dest LA)],
                                               KA,LS,dest LA)],
                                             NIL)))))],
        [([("i",1),("j",1)],um)])
    end

(*****************************************************************************) 

val wmf6 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(N,
                           OUT([PA,A,PB,B,PA,A,N],
                           INP([PS,S,PA,A],[z],
                           DEC(z,[PB,B],[z'],KA,LA2,orig LS2,
                           DEC(z',[N],[zK],pair,LAp,orig LBp,
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],zK,LA3,dest LB3)],
                           LEAK(leakA,[("i",1)]))))))))))          
               val Bwmf = PAR_I ("j",1,
                          PAR_I ("i",~1,
                          INP([PA,A,PB,B,PA,A],[xN],
                          NEW(K,
                          OUT([PB,B,PS,S,
                                ENC([PA,A,PB,B,ENC([xN,K],pair,LBp,dest LAp)],
                                    KB,LB1,dest LS1)],
                          INP([PA,A,PB,B],[u],
                          DEC(u,[zero],[um],K,LB3,orig LA3,
                          LEAK(leakB,[("j",1)]))))))))
               val Swmf = PAR_I ("j",0,
                          INP([PB,B,PS,S],[y],
                          PAR_I ("i",0,
                          DEC(y,[PA,A,PB,B],[y'],KB,LS1,orig LB1,
                          OUT([PS,S,PA,A,ENC([PB,B,y'],KA,LS2,dest LA2)],
                          LEAK(leakS,[("i",1),("j",2)]))))))
         in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
         end
    in ("wmf6","Wide Mouthed Frog (with nonces and pair/HRN)",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
         ("old session key, old nonce and all encryptions",      
              (NIL,NIL,
               NEW(oK,NEW(oN,
                   OUT([PS,S,P0,B,oK,oN],
                   OUT([PS,S,P0,B,
                        ENC([PA,A,PB,B,ENC([oN,oK],pair,LBp,dest LAp)],                                               KB,LB1,dest LS1)],
                   OUT([PS,S,P0,B,
                        ENC([PB,B,ENC([oN,oK],pair,LBp,dest LAp)],
                                               KA,LS2,dest LA2)],
                   OUT([PS,S,P0,B,ENC([zero,om],oK,LA3,dest LB3)],
                        NIL)))))))),*)
         ("old session key and tickets",      
              (NIL,NIL,
               NEW(oK,NEW(oN,
                   OUT([PS,S,P0,B,oK],
                   OUT([PS,S,P0,B,
                        ENC([PA,A,PB,B,ENC([oN,oK],pair,LBp,dest LAp)],
                                               KB,LB1,dest LS1)],
                   OUT([PS,S,P0,B,
                        ENC([PB,B,ENC([oN,oK],pair,LBp,dest LAp)],
                                               KA,LS2,dest LA2)],
                        NIL)))))))],
        [([("i",1),("j",1)],um)])
    end

(*****************************************************************************) 
(* initiator not encrypted                                                   *)
(* - introduces a flaw                                                       *)
(*****************************************************************************) 

val x   = VAR  ("x",  ["i"])
val y   = VAR  ("y",  ["i","j"])

val wmf7 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_X ("j",1,"i",
                           NEW(K,
                           OUT([PA,A,PS,S,PA,A,PB,B,ENC([K],KA,LA1,dest LS1)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],K,LA3,dest LB3)], 
                           LEAK(leakA,[("i",1)])))))))            
                val Bwmf = PAR_I("j",1,
                           PAR_I("i",~1,
                           INP([PS,S,PB,B,PA,A],[y], 
                           DEC(y,[],[yK],KB,LB2,orig LS2,
                           INP([PA,A,PB,B],[z],
                           DEC(z,[zero],[zm],yK,LB3,orig LA3,
                           LEAK(leakB,[("j",1)])))))))
                val Swmf = PAR_I("i",0,
		           PAR_I("j",0,
                           INP([PA,A,PS,S,PA,A,PB,B],[x],                            
                           DEC(x,[],[xK],KA,LS1,orig LA1,
                           OUT([PS,S,PB,B,PA,A,ENC([xK],KB,LS2,dest LB2)],
                           LEAK(leakS,[("i",1),("j",2)]))))))
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("wmf7","Wide Mouthed Frog (with initiator and responder flaw)",spec,
        [] : (string * (Proc * Proc * Proc)) list,
        [([("i",1),("j",1)],zm)])
    end 


(*****************************************************************************) 
(* Experiment for ForLysa... indexed parallel at outer level                 *)
(*****************************************************************************) 


val wmf11 =
    let fun spec (leakA,leakB,leakS) =
            let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
                val Awmf = PAR_I ("i",1,
                           PAR_I ("j",1,
                           NEW(K,
                           OUT([PA,A,PS,S,PA,A,ENC([PB,B,K],KA,LA1,dest LS1)], 
                           NEW(msg,
                           OUT([PA,A,PB,B,ENC([zero,msg],K,LA3,dest LB3)], 
                           LEAK(leakA,[("i",1)])))))))            
                val Bwmf = PAR_I("j",1,
		           PAR_I("i",~1,
                           INP([PS,S,PB,B],[y], 
                           DEC(y,[PA,A],[yK],KB,LB2,orig LS2,
                           INP([PA,A,PB,B],[z],
                           DEC(z,[zero],[zm],yK,LB3,orig LA3,
                           LEAK(leakB,[("j",1)])))))))
                val Swmf = PAR_I("i",0,				  
                           PAR_I("j",0,
                           INP([PA,A,PS,S,PA,A],[VAR("x",["i","j"])], 
                           DEC(VAR("x",["i","j"]),[PB,B],[xK],KA,LS1,orig LA1,
                           OUT([PS,S,PB,B,ENC([PA,A,xK],KB,LS2,dest LB2)],
                           LEAK(leakS,[("i",1),("j",2)]))))))
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
    in ("wmf11","Wide Mouthed Frog - indexed papallel on outer level",spec,
        [(*("session key leaked by S", (NIL,NIL,OUT([PS,S,P0,B,xK],NIL))),
         ("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
         ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),*)
         ("old session key and tickets",  
               (NIL,NIL,
                 NEW(oK,
                   OUT([PS,S,P0,B,oK,
                        ENC([PB,B,oK],KA,LA1,dest LS1),
                        ENC([PA,A,oK],KB,LS2,dest LB2)],
                   NIL))))],
        [([("i",~1),("j",~1)],zm),
         ([("i",~1),("j",~1)],yK)])
    end 


end
end
