(*****************************************************************************) 
(* Yahalom protocol                                                          *)
(*****************************************************************************) 
structure Yahalom =
struct
local
    open Lysa
    open Protocols
in

val NA  = NAME ("N^A",["i","j"])
val NB  = NAME ("N^B",["i","j"])
val x   = VAR  ("x",  ["i","j"])
val y   = VAR  ("y",  ["j"])
val yA  = VAR  ("y^A",["i","j"])
val yB  = VAR  ("y^B",["i","j"])

val zA  = VAR  ("z^A",  ["i","j"])
val zB  = VAR  ("z^B",  ["i","j"])
val zN  = VAR  ("z^N",  ["i","j"])
val zK  = VAR  ("z^K",  ["i","j"])
val vB  = VAR  ("v^B",  ["i","j"])
val v   = VAR  ("v",  ["i","j"])
val vK  = VAR  ("v^K",  ["i","j"])

val w   = VAR  ("w",  ["i","j"])
val wm  = VAR  ("w^m",["i","j"])

val oK  = NAME ("oK_{12}",[])
val oN  = NAME ("oN_{12}",[])
val oNA = NAME ("oN^A_{12}",[])
val oNB = NAME ("oN^B_{12}",[])
val om  = NAME ("om_{12}",[])


val LA'  = ("A'", ["i"])
val LB'  = ("B'", ["j"])

(*****************************************************************************) 
(*****************************************************************************) 

val y1 =
    let fun spec (leakA,leakB,leakS) =
         let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
             val Awmf = PAR_I ("i",1,
                        PAR_X ("j",1,"i",
                        NEW(NA,
                        OUT([PA,A,PB,B,PA,A,NA], 
                        INP([PS,S,PA,A],[zA,zB],
                        DEC(zA,[PB,B,NA],[zK,zN],KA,LA3,orig LS3,
                        OUT([PA,A,PB,B,zB,ENC([zN],zK,LA5,dest LB5)],
                        NEW(msg,
                        OUT([PA,A,PB,B,ENC([zero,msg],zK,LA6,dest LB6)], 
                        LEAK(leakA,[("i",1)]))))))))))           
             val Bwmf = PAR_I("j",1,
                        PAR_I("i",~1,
                        INP([PA,A,PB,B,PA,A],[x],
                        NEW(NB, 
                        OUT([PB,B,PS,S,PB,B,ENC([PA,A,x,NB],KB,LB2,dest LS2)],
                        INP([PA,A,PB,B],[vB,v],
                        DEC(vB,[PA,A],[vK],KB,LB4,orig LS4,
                        DEC(v,[NB],[],vK,LB5,orig LA5,
                        INP([PA,A,PB,B],[w],
                        DEC(w,[zero],[wm],vK,LB6,orig LA6,
                        LEAK(leakB,[("j",1)])))))))))))
             val Swmf = PAR_I("j",0,
                        INP([PB,B,PS,S,PB,B],[y],
                        PAR_I("i",0,
                        DEC(y,[PA,A],[yA,yB],KB,LS2,orig LB2,
                        NEW(K,
                        OUT([PS,S,PA,A,ENC([PB,B,yA,K,yB],KA,LS3,dest LA3),
                                 ENC([PA,A,K],KB,LS4,dest LB4)],
                        LEAK(leakS,[("i",1),("j",2)]))))))) 
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
           in ("y1","Yahalom protocol",spec,
              [("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
               ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
               ("old session key, old A- and B-nonces and all encryptions",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,NEW(om,
                             OUT([PS,S,P0,B,oK,oNA,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNA,oNB],KB,LB2,dest LS2)],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK,oNB],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oK],KB,LS4,dest LB4)],
                             OUT([PS,S,P0,A,
                                  ENC([oNB],oK,LA5, dest LB5)],
                             OUT([PS,S,P0,A,
                                  ENC([zero,om],oK,LS6, dest LB6)],
                             NIL)))))))))))),
               ("old session key and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oK],
                             OUT([PS,S,P0,A,                                  ENC([PB,B,oNA,oK,oNB],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oK],KB,LS4,dest LB4)],
                             NIL)))))))),
               ("old B-nonce and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK,oNB],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oK],KB,LS4,dest LB4)],
                             NIL))))))))],
              [([("i",1),("j",1)],wm)])
           end 


(*****************************************************************************) 
(*****************************************************************************) 

val yB  = VAR  ("y^B",["j"])

val y2 =
    let fun spec (leakA,leakB,leakS) =
         let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
             val Awmf = PAR_I ("i",1,
                        PAR_X ("j",1,"i", 
                        NEW(NA,
                        OUT([PA,A,PA,B,PA,A,NA], 
                        INP([PS,S,PA,A],[zA,zB],
                        DEC(zA,[PA,B,NA],[zK,zN],KA,LA3,orig LS3,
                        OUT([PA,A,PB,B,zB,ENC([zN],zK,LA5,dest LB5)],
                        NEW(msg,
                        OUT([PA,A,PB,B,ENC([zero,msg],zK,LA6,dest LB6)], 
                        LEAK(leakA,[("i",1)]))))))))))           
             val Bwmf = PAR_I("j",1,
                        PAR_I("i",~1,
                        INP([PA,A,PB,B,PA,A],[x],
                        NEW(NB, 
                        OUT([PB,B,PS,S,PB,B,NB,ENC([PA,A,x],KB,LB2,dest LS2)],
                        INP([PA,A,PB,B],[vB,v],
                        DEC(vB,[PA,A],[vK],KB,LB4,orig LS4,
                        DEC(v,[NB],[],vK,LB5,orig LA5,
                        INP([PA,A,PB,B],[w],
                        DEC(w,[zero],[wm],vK,LB6,orig LA6,
                        LEAK(leakB,[("j",1)])))))))))))
             val Swmf = PAR_I("j",0,
                        INP([PB,B,PS,S,PB,B],[yB,y],
                        PAR_I("i",0,
                        DEC(y,[PA,A],[yA],KB,LS2,orig LB2,
                        NEW(K,
                        OUT([PA,S,PA,A,ENC([PB,B,yA,K,yB],KA,LS4,dest LA4),
                                 ENC([PA,A,K],KB,LS3,dest LB3)],
                        LEAK(leakS,[("i",1),("j",2)]))))))) 
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
           in ("y2","Yahalom protocol (flawed)",spec,
              [("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
               ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
               ("old session key, old A- and B-nonces and all encryptions",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oK,oNA,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNA],KB,LB2,dest LS2)],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK,oNB],KA,LS4,dest LA4)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oK],KB,LS3,dest LB3)],
                             NIL))))))))),
               ("old session key and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oK],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK,oNB],KA,LS4,dest LA4)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oK],KB,LS3,dest LB3)],
                             NIL)))))))),
               ("old B-nonce and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK,oNB],KA,LS4,dest LA4)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oK],KB,LS3,dest LB3)],
                             NIL))))))))],
              [([("i",1),("j",1)],wm)])
           end 

(*****************************************************************************) 

val y3 =
    let fun spec (leakA,leakB,leakS) =
         let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
             val Awmf = PAR_I ("i",1,
                        PAR_X ("j",1,"i", 
                        NEW(NA,
                        OUT([PA,A,PB,B,A,NA], 
                        INP([PS,S,PA,A],[zN,zA,zB],
                        DEC(zA,[PB,B,NA],[zK],KA,LA3,orig LS3,
                        OUT([PA,A,PB,B,zB,ENC([zN],zK,LA5,dest LB5)],
                        NEW(msg,
                        OUT([PA,A,PB,B,ENC([zero,msg],zK,LA6,dest LB6)], 
                        LEAK(leakA,[("i",1)]))))))))))           
             val Bwmf = PAR_I("j",1,
                        PAR_I("i",~1,
                        INP([PA,A,PB,B,PA,A],[x],
                        NEW(NB, 
                        OUT([PB,B,PS,S,PB,B,NB,ENC([PA,A,x],KB,LB2,dest LS2)],
                        INP([PA,A,PB,B],[vB,v],
                        DEC(vB,[PA,A,PB,B,NB],[vK],KB,LB4,orig LS4,
                        DEC(v,[NB],[],vK,LB5,orig LA5,
                        INP([PA,A,PB,B],[w],
                        DEC(w,[zero],[wm],vK,LB6,orig LA6,
                        LEAK(leakB,[("j",1)])))))))))))
             val Swmf = PAR_I("j",0,
                        INP([PB,B,PS,S,PB,B],[yB,y],
                        PAR_I("i",0,
                        DEC(y,[PA,A],[yA],KB,LS2,orig LB2,
                        NEW(K,
                        OUT([PS,S,PA,A,yB,ENC([PB,B,yA,K],KA,LS3,dest LA3),
                                    ENC([PA,A,PB,B,yB,K],KB,LS4,dest LB4)],
                        LEAK(leakS,[("i",1),("j",2)]))))))) 
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
           in ("y3","Yahalom protocol (BAN opt, Paulson fix)",spec,
              [("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
               ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
               ("old session key, old A- and B-nonces and all encryptions",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,NEW(om,
                             OUT([PS,S,P0,B,oK,oNA,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNA,oNB],KB,LB2,dest LS2)],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,PB,B,oNB,oK],KB,LS4,dest LB4)],
                             OUT([PS,S,P0,A,
                                  ENC([oNB],oK,LA5,dest LB5)],
                             OUT([PS,S,P0,A,
                                  ENC([zero,om],KB,LA6,dest LB6)],
                             NIL)))))))))))),
               ("old session key and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oK],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,PB,B,oNB,oK],KB,LS4,dest LB4)],
                             NIL)))))))),
               ("old B-nonce and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,PB,B,oNB,oK],KB,LS4,dest LB4)],
                             NIL))))))))],
              [([("i",1),("j",1)],wm)])
           end 

(*****************************************************************************) 

val y4 =
    let fun spec (leakA,leakB,leakS) =
         let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
             val Awmf = PAR_I ("i",1,
                        PAR_X ("j",1,"i",
                        NEW(NA,
                        OUT([PA,A,PB,B,PA,A,NA], 
                        INP([PS,S,PA,A],[zN,zA,zB],
                        DEC(zA,[PB,B,NA],[zK],KA,LA3,orig LS3,
                        OUT([PA,A,PB,B,zB,ENC([zN],zK,LA5,dest LB5)],
                        NEW(msg,
                        OUT([PA,A,PB,B,ENC([zero,msg],zK,LA6,dest LB6)], 
                        LEAK(leakA,[("i",1)]))))))))))           
             val Bwmf = PAR_I("j",1,
                        PAR_I("i",~1,
                        INP([PA,A,PB,B,PA,A],[x],
                        NEW(NB, 
                        OUT([PB,B,PS,S,PB,B,NB,ENC([PA,A,x],KB,LB2,dest LS2)],
                        INP([PA,A,PB,B],[vB,v],
                        DEC(vB,[PA,A,NB],[vK],KB,LB4,orig LS4,
                        DEC(v,[NB],[],vK,LB5,orig LA5,
                        INP([PA,A,PB,B],[w],
                        DEC(w,[zero],[wm],vK,LB6,orig LA6,
                        LEAK(leakB,[("j",1)])))))))))))
             val Swmf = PAR_I("j",0,
                        INP([PB,B,PS,S,PB,B],[yB,y],
                        PAR_I("i",0,
                        DEC(y,[PA,A],[yA],KB,LS2,orig LB2,
                        NEW(K,
                        OUT([PS,S,PA,A,yB,ENC([PB,B,yA,K],KA,LS3,dest LA3),
                                    ENC([PA,A,yB,K],KB,LS4,dest LB4)],
                        LEAK(leakS,[("i",1),("j",2)]))))))) 
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
           in ("y4","Yahalom protocol (BAN opt)",spec,
              [("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
               ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
               ("old session key, old A- and B-nonces and all encryptions",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,NEW(om,
                             OUT([PS,S,P0,B,oK,oNA,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNA,oNB],KB,LB2,dest LS2)],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNB,oK],KB,LS4,dest LB4)],
                             OUT([PS,S,P0,A,
                                  ENC([oNB],oK,LA5,dest LB5)],
                             OUT([PS,S,P0,A,
                                  ENC([zero,om],KB,LS6,dest LB6)],
                             NIL)))))))))))),
               ("old session key and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oK],
                             OUT([PS,S,P0,A,                                  ENC([PB,B,oNA,oK],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNB,oK],KB,LS4,dest LB4)],
                             NIL)))))))),
               ("old B-nonce and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS3,dest LA3)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNB,oK],KB,LS4,dest LB4)],
                             NIL))))))))],
              [([("i",1),("j",1)],wm)])
           end 

val y4a =
    let fun spec (leakA,leakB,leakS) =
         let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
             val Awmf = PAR_I ("i",1,
                        PAR_X ("j",1,"i",
                        NEW(NA,
                        OUT([PA,A,PB,B,PA,A,NA], 
                        INP([PS,S,PA,A],[zN,zA,zB],
                        DEC(zA,[PB,B,NA],[zK],KA,LA,orig LS,
                        OUT([PA,A,PB,B,zB,ENC([zN],zK,LA,dest LB)],
                        NEW(msg,
                        OUT([PA,A,PB,B,ENC([zero,msg],zK,LA,dest LB)], 
                        LEAK(leakA,[("i",1)]))))))))))           
             val Bwmf = PAR_I("j",1,
                        PAR_I("i",~1,
                        INP([PA,A,PB,B,PA,A],[x],
                        NEW(NB, 
                        OUT([PB,B,PS,S,PB,B,NB,ENC([PA,A,x],KB,LB,dest LS)],
                        INP([PA,A,PB,B],[vB,v],
                        DEC(vB,[PA,A,NB],[vK],KB,LB,orig LS,
                        DEC(v,[NB],[],vK,LB,orig LA,
                        INP([PA,A,PB,B],[w],
                        DEC(w,[zero],[wm],vK,LB,orig LA,
                        LEAK(leakB,[("j",1)])))))))))))
             val Swmf = PAR_I("j",0,
                        INP([PB,B,PS,S,PB,B],[yB,y],
                        PAR_I("i",0,
                        DEC(y,[PA,A],[yA],KB,LS,orig LB,
                        NEW(K,
                        OUT([PS,S,PA,A,yB,ENC([PB,B,yA,K],KA,LS,dest LA),
                                    ENC([PA,A,yB,K],KB,LS,dest LB)],
                        LEAK(leakS,[("i",1),("j",2)]))))))) 
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
           in ("y4a","Yahalom protocol (BAN opt)",spec,
              [("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
               ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
               ("old session key, old A- and B-nonces and all encryptions",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,NEW(om,
                             OUT([PS,S,P0,B,oK,oNA,oNB],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNA,oNB],KB,LB,dest LS)],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS,dest LA)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNB,oK],KB,LS,dest LB)],
                             OUT([PS,S,P0,A,
                                  ENC([oNB],oK,LA,dest LB)],
                             OUT([PS,S,P0,A,
                                  ENC([zero,om],KB,LS,dest LB)],
                             NIL)))))))))))),
               ("old session key and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oK],
                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS,dest LA)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNB,oK],KB,LS,dest LB)],
                             NIL)))))))),
               ("old B-nonce and tickets",  
                      (NIL,NIL, 
                          NEW(oK,NEW(oNA,NEW(oNB,
                             OUT([PS,S,P0,B,oNB],                             OUT([PS,S,P0,A,
                                  ENC([PB,B,oNA,oK],KA,LS,dest LA)],
                             OUT([PS,S,P0,A,
                                  ENC([PA,A,oNB,oK],KB,LS,dest LB)],
                             NIL))))))))],
              [([("i",1),("j",1)],wm)])
           end 

(*****************************************************************************) 
 

(*****************************************************************************) 
 

val y5 =
    let fun spec (leakA,leakB,leakS) =
         let val Nwmf = [([("i",1)],PA),([("i",1)],KA),([("j",1)],KB)]
             val Awmf = PAR_I ("i",1,
                        PAR_X ("j",1,"i",
                        NEW(NA,
                        OUT([PA,A,PB,B,PA,A,NA], 
                        INP([PS,S,PA,A],[zA,zB],
                        DEC(zA,[PB,B,NA],[zK,zN],KA,LA,orig LS,
                        OUT([PA,A,PB,B,zB,ENC([zN],zK,LA',dest LB')],
                        NEW(msg,
                        OUT([PA,A,PB,B,ENC([zero,msg],zK,LA,dest LB)], 
                        LEAK(leakA,[("i",1)]))))))))))           
             val Bwmf = PAR_I("j",1,
                        PAR_I("i",~1,
                        INP([PA,A,PB,B,PA,A],[x],
                        NEW(NB, 
                        OUT([PB,B,PS,S,PB,B,ENC([PA,A,x,NB],KB,LB,dest LS)],
                        INP([PA,A,PB,B],[vB,v],
                        DEC(vB,[three,PA,A,NB],[vK],KB,LB,orig LS,
                        DEC(v,[NB],[],vK,LB',orig LA',
                        INP([PA,A,PB,B],[w],
                        DEC(w,[zero],[wm],vK,LB,orig LA,
                        LEAK(leakB,[("j",1)])))))))))))
             val Swmf = PAR_I("j",0,
                        INP([PB,B,PS,S,PB,B],[y],
                        PAR_I("i",0,
                        DEC(y,[PA,A],[yA,yB],KB,LS,orig LB,
                        NEW(K,
                        OUT([PS,S,PA,A,ENC([PB,B,yA,K,yB],KA,LS,dest LA),
                                 ENC([three,PA,A,yB,K],KB,LS,dest LB)],
                        LEAK(leakS,[("i",1),("j",2)]))))))) 
             in NEW_I(Nwmf,BANG(PAR [Awmf,Bwmf,Swmf]))
             end
           in ("y5","Yahalom protocol (as y1, but $N^B_i$ inside ticket from $S$))",spec,
              [("master key $K^A_1$",  (OUT([PA,A,P0,B,KA],NIL), NIL, NIL)),
               ("master key $K^B_1$",  (NIL, OUT([PB,B,P0,B,KB],NIL), NIL)),
               ("old session key $oK_{12}$",  
                      (NIL,NIL, NEW(oK,NEW(oNA,NEW(oNB,
                                        OUT([PS,S,P0,B,oK,
                                           ENC([PA,oNA,oNB],KB,LB,dest LS),
                                           ENC([PB,B,oNA,oK,oNB],KA,LS,dest LA),
                                           ENC([three,PA,A,oNB,oK],KB,LS,dest LB)],
                                          NIL))))))],
              [([("i",1),("j",1)],wm)])
           end 
end
end