structure BreachFunction = 
struct

local
open LysaN
in
(*

There are two different breach functions. Authentication and secrecy.
Authentication-breach-function determies whether a given breach can take place, 
based on:
The breach
a substitution
a set of constraints or the corresponding mismatch list

Secrecy-breach-function determies whether a given breach can take place, 
based on:
The breach
a substitution
The final DY-knowledge

*)


datatype Breach = AUT of Term*Term | SEC of Term

(* t is the breach term, sub is a substitution, DY is the DY knowledge *)

fun secrecyBreach(t,sub,DY) = let val DY' = EvalConstraints.applySubstitutionTs sub DY
                              in DyKnowledge.elementIn(t,DY') end

fun authenticationBreach((orig,dest),((cp1,cp1'),(cp2,cp2'))::mismatch) = 
          (orig,dest) = (cp2',cp1) orelse authenticationBreach((orig,dest),mismatch)
  | authenticationBreach(_,[]) = false

fun breach(breach,sub,mismatch,DY) = 
       case breach of
         AUT(orig,dest) => authenticationBreach((orig,dest),mismatch)
        |SEC(t)         => secrecyBreach(t,sub,DY)
 
end
end            