structure EvalConstraints =
struct

local 
open Types
open LysaN
open Constraints
open ToString
val I = CNAME("I",0)


val lab = ref 0;
fun nextLabel() = let val _ = lab := !lab + 1
                      val _ = if !lab mod 100 = 0 then print(".") else ()
                  in !lab end
fun resetLabel() = lab := 0

val debug = false
val debug'' = false
fun print'(str) = if debug then print(str) else ()
fun print''(str) = if debug'' then print(str) else ()

in
(*

type substitution = (Term * Term) list

p  ::= n | x | %x | E(p1,...,pk)p0, p0 or subpatterns of p0 cannot contain %x. fv(p) must be non-empty.
t  ::= n | x | E(t1,...,tk)t0


constraint ::= (p,[t1,...,tk])
singleSubstitution ::= (%x,t)
substitution ::= singleSubstitution list

-----------------------------------------------------------------------------
Cases:
1: (%x',[terms]),substitution
constraint pattern is a variable

3: ({p1,...,pk}p0(d<-o),[terms]),substitution
constraint pattern is an encryption

Substitute variables in encryption according to substitution. Check whether the encryption can be generated directly 



*)
infix ==


fun applySubstitutionToConstraint sub (t,IK) = (applySubstitution sub t,applySubstitutionTs sub IK)

(* sub is applyed to all mismatches of psi. The once that are no longer mismatches are removed from the list *)
fun applySubstitutionToPsi sub (((t1,t1'),(t2,t2'))::psi) = let val t1 = applySubstitution sub t1
                                                                val t1' = applySubstitution sub t1'
                                                                val t2 = applySubstitution sub t2
                                                                val t2' = applySubstitution sub t2'
                                                            in if t1 <> t1' orelse t2 <> t2' 
                                                               then ((t1,t1'),(t2,t2'))::(applySubstitutionToPsi sub (psi))
                                                               else applySubstitutionToPsi sub (psi)
                                                            end
  | applySubstitutionToPsi sub [] = []

(*
(* constraint * substitution -> bool *)
fun constraintSat((t,ts),sub) = let val t' = applySubstitution sub t in DY.elementIn(t',ts) end

fun constraintSetSat(c::cs,sub) = let val (b1,mismatch1) = constraintSat(c,sub)
                                      val (b2,mismatch2) = if b1 then constraintSetSat(cs,sub) else (false,[])
                                  in (b1 andalso b2,if (b1 andalso b2) then mismatch1@mismatch2 else []) end
  | constraintSetSat([],_) = (true,[])

*)
type substitution = (CTerm * CTerm) list
type mismatch = (CTerm * CTerm)*(CTerm * CTerm)
datatype state = State of {labels: int list,cs:constraint list, scs:constraint list, sub:substitution, psi:mismatch list} (* cs is untreated constraints. scs is simple constraints which have been looked at. sub is the substitution so far *)
datatype pathCs = PathCs of {path:Graph.path, statelist: state list, termList:string list}


(*
datatype result = SOLLUTION of substitution*mismatch*path*constraint list
                | POTENTIALPATHS of path list
*)
(* a state can be transformed into a number of states using the various transform rules on the various constraints 

 Analysing an encryption constraint. Two possibilities:
  1. We can make a direct substitution. The constrain disappears and the substitution is extended.
  2. We can split it into a number of new constraints
 
 Every time we deal with an encryption constraint, we have a branching. At leaves of the tree we are left with
 simple constraints and substitutions. Can these simple constraints be satisfied under the corresponding substitutions?
 If the can, what errorconditions do they lead to?

*)


fun addMismatch((cp1,CORIG(cp2)),(cp2',CDEST(cp1')),mismatches) = if not(cp1 == cp1') orelse not(cp2  == cp2') 
                                                                  then ((cp1,cp1'),(cp2,cp2'))::mismatches 
                                                                  else mismatches
  | addMismatch (_,_,mismatches) = mismatches (*raise Problem("Orig Dest")*)

(* The function creates a substitution, which unifies the two terms if possible otherwise it retrns NONE *)
fun unifier'((t1 as CVAR(_),t2 as CVAR(_)),psi) = if t1 == t2 then SOME([],psi) else SOME([(t1,t2)],psi)
  | unifier'((t1 as CVAR(_),t2),psi) = SOME([(t1,t2)],psi)
  | unifier'((t1,t2 as CVAR(_)),psi) = SOME([(t2,t1)],psi) (*er det ikke rigtigt?*)
  | unifier'((t1 as CNAME(_),t2 as CNAME(_)),psi) = if t1 == t2 then SOME([],psi) else NONE (* er det ikke rigtigt?*)
  | unifier'((CKEY(t1),CKEY(t2)),psi) = unifier'((t1,t2),psi)
  | unifier'((CENC(ts1,t1,cp1,_),CENC(ts2,t2,cp2,_)),psi) = let val psi' = addMismatch(cp1,cp2,psi)
                                                           in unifierLs((t1::ts1,t2::ts2),psi') end
  | unifier' _ = NONE


and unifierLs((t1::ts1,t2::ts2),psi) = (case unifier'((t1,t2),psi) of 
                                          NONE      => NONE 
                                         |SOME(ls1,psi1) => (
                                            case unifierLs((ts1,ts2),psi) of
                                              NONE      => NONE
                                             |SOME(ls2,psi2) => SOME(ls1@ls2,psi1@psi2)
                                            )
                                       )
  | unifierLs(([],[]),psi) = SOME([],psi)
  | unifierLs _ = NONE


fun unifier((t1 as CKEY(_),t2 as CKEY(_)),psi) = unifier'((t1,t2),psi)
  | unifier((t1 as CENC(_),t2 as CENC(_)),psi) = unifier'((t1,t2),psi)
  | unifier _ = NONE

fun generateAllUnifiers(_,[]) = []
  | generateAllUnifiers(t,t'::IK) = case unifier((t,t'),[]) of
                                      NONE => generateAllUnifiers(t,IK)
                                     |SOME(sub,psi) => (sub,psi)::generateAllUnifiers(t,IK) (*for each generated substitution we also generate the corresponding psi component *)

fun generateConstraintsLs(t::ts,IK) = (t,IK) :: generateConstraintsLs(ts,IK)
  | generateConstraintsLs ([],_) = []

fun generateConstraints(CENC(ts,t,_,_),IK) = (t,IK)::generateConstraintsLs(ts,IK)
  | generateConstraints _ = raise Problem("")

datatype constraintType = ENCRYPTION of (CTerm*CTerm) | SIMPLE | CONSTANT | KEYC  (*encryption type contains cryptopoint information (sender,intended reciever)*)

fun constraintType(CENC(_,_,(t1,CORIG(t2)),_),_) = ENCRYPTION(t1,t2)
  | constraintType(CENC(_,_,(t1,CDEST(t2)),_),_) = ENCRYPTION(t1,t2)(*raise Problem("DEST should not occur in LHS of constraint")*)
  | constraintType(CVAR(_),_)     = SIMPLE
  | constraintType(CNAME(_),_)    = CONSTANT
  | constraintType(CKEY(t),_)     = KEYC



(* a member function which ignores the flags of the terms in IK *)
fun memberOfIK(t,IK) = List.exists (fn t' => t == t') IK

(* adds a termlist to a termlist. Terms that allready exists are not added. Returns true if changes where made *)
fun addToIK(t::ts,IK) = let val (IK',change) = addToIK(ts,IK)
                        in if memberOfIK(t,IK') then (IK',change) else (t::IK',true) end
  | addToIK([],IK) = (IK,false)

(* removes a term t from IK. Ignoring the flags *)
fun removeFromIK(t,IK) = List.filter (fn t' => not(t == t')) IK

(* used to determine whether a constant constraint is satisfiable *)
fun satisfiable(t as CNAME(_),IK) = memberOfIK(t,IK)
  | satisfiable _     = raise Problem("Satisfiable should only be used with constant constraints")

(*marks the term encT in IK as "treated" *)
fun markTermInIK(t as CENC(ts,key,cps,_),IK) = List.map (fn t' => if t == t' then CENC(ts,key,cps,true) else t') IK
  | markTermInIK _ = raise Problem("can only mark encryption terms")

fun markTermInCs(encT as CENC(_),cs) = List.map (fn (t,IK) => (t,markTermInIK(encT,IK))) cs

(* marks the term encT in IK as "treated" and adds the constent, ts, of the encryption term to IK *)
fun decryptTermInIK(encT as CENC(ts,key,_,_),IK) = 
  if memberOfIK(encT,IK) 
  then let val (IK',_) = addToIK(ts,markTermInIK(encT,IK)) in IK' end 
  else IK
  | decryptTermInIK _ = raise Problem ""

fun decryptTermInCs(encT,cs) = List.map (fn (t,IK) => (t,decryptTermInIK(encT,IK))) cs


fun simpleDecryptIK'((t as CENC(ts,key,(cp2',CDEST(cp1')),b))::tsRest,IK,change,psi) = 
  if not b andalso memberOfIK(key,IK) (* The term has not been decrypted and the key is actually a member of IK*)
  then let val (IK',change') = addToIK(ts,IK)  (* the decrypted terms are added to IK *)
           val IK'' = markTermInIK(t,IK')      (* the decrypted encryption term is marked. We don't have to consider the corresponding terms in other constraints since this is simple decryption. If it is possible here it will also be possible and hence be performed in other constraints*)
           val _ = print'("decrypting "^cTermToString(t)^"\n")
       in simpleDecryptIK'(tsRest,IK'',change orelse change',((I,cp1'),(cp2',cp2'))::psi) end
  else simpleDecryptIK'(tsRest,IK,change,psi)
  | simpleDecryptIK'((t as CENC(_))::tsRest,_,_,_) = (print ("**"^cTermListToString(t::tsRest));raise Problem("ORIG should not occur on RHS of constraint"))
  | simpleDecryptIK'(t::tsRest,IK,change,psi) = simpleDecryptIK'(tsRest,IK,change,psi)
  | simpleDecryptIK'([],IK,change,psi) = if change then simpleDecryptIK'(IK,IK,false,psi) else (IK,psi)

(* decrypts whatever can be decrypted in IK. If terms are succesfully decrypted, the process is iterated. We only decrypt terms where the key does not require further analysis *)
fun simpleDecryptIK(IK) = let val _ = print' "simpleDecrypt start\n"
                              val (IK,psi) = simpleDecryptIK'(IK,IK,false,[])
                              val _ = print' "simpleDecrypt end\n"
                          in (IK,psi) end

fun notSimple(CNAME _) = false
  | notSimple(CKEY(CNAME _)) = false
  | notSimple _ = true

(* returns the first encryption term which has not been treated yet with a non-simple key. Otherwise NONE *)
fun advancedDecryptIK(IK) = List.find (fn CENC(ts,key,cps,b) => not b andalso notSimple(key)|_ => false) IK

fun stateIsFinal(State{cs,...}) = cs = []

val substitutionToString = ToString.listToString (fn (t1,t2)=> (cTermToString t1)^"->"^(cTermToString t2))

(* state to string *)
fun stateToString(State{labels,cs,scs,sub,psi}) =
                                           (if debug then "label:"^(ToString.listToString Int.toString labels)^"\n" else "")^
                                           "state:\n"^
                                           "vcs =\n"^constraintSetToString(scs)^"\n"^
                                           (if cs = [] then "" else "cs =\n"^constraintSetToString(cs)^"\n")^
                                           "sub =\n"^substitutionToString(sub)^"\n"^
                                           "psi =\n"^psiToString(psi)^"\n"

fun stateListToString'(s::ss) = stateToString(s)^stateListToString'(ss)
  | stateListToString' [] = ""
fun stateListToString(ss) = "**state list:** ("^Int.toString(List.length(ss))^")\n"^stateListToString'(ss)^"***************\n"


(*The substitution is applied to all constraints in scs and cs. Any constraints that are no longer simple are moved to cs.*)
(* apply substitution and filter the scs. Substitution is also applied to psi *)
fun normalizeState(s as State{cs,scs,sub,psi,labels}) =
  let val cs' = List.map (applySubstitutionToConstraint sub) cs
      val scs' = List.map (applySubstitutionToConstraint sub) scs
      val (scs'',notSimple) = List.partition (fn c => constraintType(c) = SIMPLE) scs'
      val psi' = applySubstitutionToPsi sub psi
      val s' = State{cs = rev(notSimple)@cs',scs = scs'',sub=sub,psi=psi',labels=nextLabel()::labels}(* can we just add the constraints that are no longer simple to the front of the list cs? *)
      val _ = print'("Normalizing state:\n"^(stateToString(s))^"\n") (*"Normalizing state:\n"^(stateToString(State{cs,scs,sub,psi,labels}))^"\n"*)
      val _ = print'("state normalized:\n"^stateToString(s')^"\n")
  in s' end

(* this function is called recursively as long as we have untreated encryption terms. The result is a list of states, where the first constraint in cs only contains treated encryption terms and the treated terms of the first constraint also has been treated in the previous and succesive constraints *)
(* The argument has to be a normalized state *)
fun constraintDecryption(s as State{cs,scs,sub,psi,labels}) = 
(*  if constraintType(hd(cs)) = SIMPLE then [s] else*)
  let val (t,IK) = hd(cs)
      val _ = print'("Performing constraint decryption of constraint:\n"^constraintToString (t,IK)^"\n")
      val (IK',psi') = simpleDecryptIK(IK) (* A simple decrypt is performed on IK'. This gives us IK'' and psi' *)
      val _ = print'("IK after simple decrypt:\n"^cTermListToString IK'^"\n")
  in case advancedDecryptIK(IK') of (* if there exists an untreated encryption term it is returned as SOME(t)*)
       NONE => [State{cs=(t,IK')::tl(cs),scs=scs,sub=sub,psi=psi'@psi,labels=labels}] (* there are no untreated encryption terms. We return the input state with the substitution and the simpleDecrypt performed to the next constraint in the list *)
      |SOME(encT as CENC(ts,key,(cp2',CDEST(cp1')),b)) => 
         let val _ = print''("**Decrypting the term "^cTermToString encT^"\nState:"^stateToString(s))
             val newC = (key,removeFromIK(encT,IK'))
             val IKnoDecrypt = markTermInIK(encT,IK')
             val IKdecrypt = decryptTermInIK(encT,IK')
             val csNoDecrypt = cs (*markTermInCs(encT,tl cs)*)
             val csDecrypt = decryptTermInCs(encT,tl cs)
             val scsNoDecrypt = scs (*markTermInCs(encT,scs)*)
             val scsDecrypt = scs (*decryptTermInCs(encT,scs)*)
             val stateNoDecrypt = State{cs=(t,IKnoDecrypt)::tl(csNoDecrypt),scs=scsNoDecrypt,sub=sub,psi=psi'@psi,labels=nextLabel()::labels}
             val stateDecrypt = State{cs=newC::(t,IKdecrypt)::csDecrypt,scs=scsDecrypt,sub=sub,psi=((I,cp1'),(cp2',cp2'))::psi'@psi,labels=nextLabel()::labels}
             val _ = print''("two new states from decrypt:\n"^stateListToString [stateDecrypt,stateNoDecrypt]^"\n")
         in constraintDecryption(stateNoDecrypt)@constraintDecryption(stateDecrypt) end
  end


(*  
c1 (simple)
c2 (simple)
t:{ts}k U IK3
c3
c4

1. decrypt
c1 (simple)
c2 (simple)
k:IK3
t:*{ts}k U ts U IK3
c3 with {ts}k decrypted and marked
c4 with {ts}k decrypted and marked

2. no decrypt
c1 (simple)
c2 (simple)
t:*{ts}k U IK3
c3 with {ts}k decrypted and marked
c4 with {ts}k decrypted and marked
*)


(* can only be used when s is non-final and normalized and first constraint of cs is decrypted. *)
(*
   non-final => cs is not empty 
   normalized => substitution is applied to all constraints and to psi. All constraints in scs are in fact simple
   decrypted => all encryption terms of the first constraint in cs are treated. All corresponding encryption terms in all other constraints are treated.
*)

fun nextStates(s as State{labels,cs,scs,sub,psi}) = 
                    let val c = hd(cs) (* cs has at least one element since it is not final *)
                        val _ = print'("next states called on \n"^stateToString(s)^"\n")
                    in
                      case constraintType(c) of
                        SIMPLE     => let val newState = State{labels = nextLabel()::labels,cs=tl(cs),scs = c::scs, sub = sub,psi = psi}
                                          val _ = print'("simple constraint moved:\n"^stateToString(newState)^"\n")
                                      in [newState] end
                       |CONSTANT   => if satisfiable(c)
                                      then let val newState = State{labels=nextLabel()::labels,cs=tl(cs),scs = scs, sub = sub, psi = psi}
                                               val _ = print'("satisfiable:\n"^stateToString(newState)^"\n")
                                           in [newState] end
                                      else (print'("unsatisfiable.\n");[]) (*argument state is not satisfiable. This means that the argument state has no next states indicated by returning an empty list *)
                       |ENCRYPTION(t1,t2) => let val newSubsAndPsis = generateAllUnifiers(c) (*list of pairs of (sub,psi)*)
                                                 val newConstraints = generateConstraints(c)
                                                 val newStates = State{labels=nextLabel()::labels,cs = newConstraints @ tl(cs),scs=scs,sub=sub,psi=((t1,t1),(t2,I))::psi} :: 
                                                         (List.map (
                                                            fn (newSub,newPsi) => State{labels = nextLabel()::labels,
                                                                                        cs = tl(cs),
                                                                                        scs = scs,
                                                                                        sub = newSub@sub,
                                                                                        psi = newPsi@psi}
                                                          )
                                                          newSubsAndPsis)
                                                 val _ = print'("new states from encryption:\n"^stateListToString(newStates))
                                             in newStates end
                       |KEYC => let val newSubsAndPsis = generateAllUnifiers(c)
                                
                                    val newStates = (List.map (
                                        fn (newSub,newPsi) => State{ labels = nextLabel()::labels,
                                                                   cs = tl(cs),
                                                                   scs = scs,
                                                                   sub = newSub@sub,
                                                                   psi = newPsi@psi}
                                      ) 
                                      newSubsAndPsis) 
                                    val _ = print'("new states from key constraint:\n"^stateListToString(newStates))
                                in newStates end
                    end
 
(* From a list of states we calculate the corresponding list of final states using the functions:
     -normalizeState
     -constraintDecryption
     -nextStates
*)
fun trimPsi(State{labels,scs,cs,sub,psi}) = State{labels=labels,scs=scs,cs=cs,sub=sub,psi=Aux.removeDoubles(psi)}
(* FINAL STATES 
fun finalStates(s::statelist) = 
	      let val s' as State{cs,...} = normalizeState(s)
          in 
            if cs = [] (*s' is final*)
            then trimPsi(s')::finalStates(statelist)
            else let val decryptedStates = constraintDecryption(s')
(*                     val _ = print'("decrypted states start\n"^
                                   "decryption of state\n"^stateToString s'^"\nresult:\n"^stateListToString(decryptedStates)^"\n")
                     val _ = print'("decrypted states end\n")*)
                     val next = Aux.meltList(List.map nextStates decryptedStates) 
(*                     val _ = print' (stateListToString(next)) *)
                 in finalStates(next@statelist) end (*s' is not final. We calculate the next states of s'*)
          end
  | finalStates [] = []
*)
fun finalStates(s::statelist) = 
	      let val s' as State{cs,...} = normalizeState(s)
          in 
            if cs = [] (*s' is final*)
            then trimPsi(s')::finalStates(statelist)
            else let val decryptedStates = constraintDecryption(s')
(*                     val _ = print'("decrypted states start\n"^
                                   "decryption of state\n"^stateToString s'^"\nresult:\n"^stateListToString(decryptedStates)^"\n")
                     val _ = print'("decrypted states end\n")*)
                     val next = Aux.meltList(List.map nextStates decryptedStates) 
(*                     val _ = print' (stateListToString(next)) *)
                 in finalStates(next@statelist) end (*s' is not final. We calculate the next states of s'*)
          end
  | finalStates [] = []



fun addConstraint(breachC,State{labels,scs,cs,sub,psi}) = State{labels=labels,scs=scs,cs = cs @ [breachC],sub=sub,psi=psi}


val count = ref 0
fun increase() = (count := !count + 1;if (!count mod 1000) = 0 then print(Int.toString((!count) div 1000)) else ())
fun reset() = count := 0
(******** secrecy **********)
(* path*breach -> substitution option *)
(*
fun hasSecrecyBreach(path,breachTerm,leaks) = 
      let val _ = increase()
          val (cs,breachC) = createConstraintSet(path,Aux.union(Constraints.IK,leaks),breachTerm)
          val init = State{labels = [~1],cs=cs,scs=[],sub=[],psi=[]}
          val finals = finalStates [init]
      in 
        if finals = [] 
        then (NONE,false)
        else let val finals' = List.map (fn s => addConstraint(breachC,s)) finals
                 val finals'' = finalStates(finals')
             in 
               if finals'' = []
               then (NONE,true)
               else let val State{sub,psi,...} = hd(finals) in (SOME(sub,psi,cs),true) end (* the first and best substitution is chosen *)
             end
      end

fun hasSecrecyBreachLS(p::paths,paths',breachTerm,leaks) = (case hasSecrecyBreach(p,breachTerm,leaks) of
                                                (NONE,true) => hasSecrecyBreachLS(paths,p::paths',breachTerm,leaks)
                                               |(NONE,false) => hasSecrecyBreachLS(paths,paths',breachTerm,leaks)
                                               |(SOME(sub,psi,cs),_) => (SOME(sub,psi,p,cs),[])
                                              )
  | hasSecrecyBreachLS ([],paths',_,_) = (NONE,paths')




(******** authentication **********)
(*Between ((cp1,cp1'),(cp2,cp2')) and (orig,dest) we have the correspondance, orig = cp2', dest = cp1.*)



(* path*breach -> substitution option *)
fun hasAuthenticationBreach(path,(orig,dest),leaks) = 
  let val _ = increase()
      val (cs,breachC) = createConstraintSet(path,Aux.union(Constraints.IK,leaks),CVAR("z_bullet",0))
      val init = State{labels = [~1],cs=cs@[breachC],scs=[],sub=[],psi=[]}
      val finals = finalStates [init]
  in 
    case List.find (fn State{psi,...} => origDestInPsi((orig,dest),psi)) finals of
      NONE => if finals = [] then (NONE,false) else (NONE,true) (* false if the path is not potentially part of a longer path which yelds a solution. True if there is a chance *)
     |SOME(State{sub,psi,...}) => (SOME(sub,psi,cs),true) (* the first and best substitution is chosen *)
  end

fun hasAuthenticationBreachLS(p::paths,paths',(orig,dest),leaks) = (case hasAuthenticationBreach(p,(orig,dest),leaks) of
                                                (NONE,true) => hasAuthenticationBreachLS(paths,p::paths',(orig,dest),leaks)
                                               |(NONE,false) => hasAuthenticationBreachLS(paths,paths',(orig,dest),leaks)
                                               |(SOME(sub,psi,cs),_) => (SOME(sub,psi,p,cs),[])
                                              )
  | hasAuthenticationBreachLS([],paths',_,_) = (NONE,paths') (*No breach was found yet but paths' are the paths potentially worth extending *)

*)
(* state*transition -> state list *)

fun origDestInPsi((orig,dest),psi) = let val _ = print'("in orig-dest search. Searching for breach: ("^cTermToString(orig)^","^cTermToString(dest)^") in psi:\n"^psiToString(psi))
                                         val found = List.exists (fn ((cp1,cp1'),(cp2,cp2')) => orig == cp2' andalso dest == cp1) psi
                                         val _ = print'(if found then "found!\n" else "not found!\n")
                                     in found end


fun extendStateOUTSTEP(State{labels,cs = [](*the state must be final*),scs,sub,psi},ts) = 
      let val (_,IK) = hd(scs) (* final intruder knowledge *)
          val IK' = Aux.union(ts,resetCTermList(IK))
          val constraint = (dummyVar,IK')
          val state = State{labels=nextLabel()::labels,cs=[constraint],scs = scs,psi=psi,sub=sub} 
          val _ = print'("extending state with OUTSTEP. Result:\n"^stateToString(state)^"\n")
      in state end
fun extendStateINSTEP(State{labels,cs = [](*the state must be final*),scs,sub,psi},ts) = 
      let val (_,IK) = hd(scs) (* final intruder knowledge *)
          val newConstraints =  List.map (fn t => (t,IK)) ts
          val dummyConstraint = (dummyVar,IK)
          val state = State{labels=nextLabel()::labels,cs = newConstraints,scs = scs,psi=psi,sub=sub} (*dummyConstraint::scs ??*)
          val _ = print'("extending state with INSTEP. Result:\n"^stateToString(state)^"\n")
      in state end


(* when creating a new PathCs we solve the basic constraints *)
fun createNewPathCs((transition,node),PathCs{path,statelist,termList}) =
(*      let val _ = print'("Termlist:\n"^ToString.listToString (fn x => x) (termList)^"\n") in*)
      case transition of 
        OUTSTEP(ts) => let val (ts',termList') = termToCTermList(ts,termList) 
                           val statelist' =  List.map (fn state => extendStateOUTSTEP(state,ts')) statelist
                           val statelist'' = finalStates(statelist')
(*                           val _ = print'("\nout\n")
                           val _ = print'("ts'"^cTermListToString(ts')^"\n")
                           val _ = print'("termlist'"^ToString.listToString (fn x => x) (termList')^"\n") *)
                           in PathCs{path = node::path,statelist = statelist'',termList = termList'} end
       |INSTEP(ts)  => let val (ts',termList') = termToCTermList(ts,termList)
                           val statelist' =  List.map (fn state => extendStateINSTEP(state,ts')) statelist
                           val statelist'' = finalStates(statelist')
(*                           val _ = print'("\nin\n")
                           val _ = print'("ts'"^cTermListToString(ts')^"\n")
                           val _ = print'("termlist'"^ToString.listToString (fn x => x) (termList')^"\n")*)
(*          val _ = print'("1\n"^stateListToString(statelist')^"\n")
          val _ = print'("2\n"^stateListToString(statelist'')^"\n") *)
                           in PathCs{path = node::path,statelist = statelist'',termList = termList'} end
      
(*end*)
(*          val _ = print'("Path:\n"^ToString.stateToString(Graph.getState(node))^"+"^Graph.pathToString(path)^"\n")
          val _ = print'("1\n"^stateListToString(statelist)^"\n")
          val _ = print'("2\n"^stateListToString(statelist')^"\n")
          val _ = print'("3\n"^stateListToString(statelist'')^"\n")
*)
(* pathCs -> pathCs list *)
fun nextPathCs(P as PathCs{path,...}) = 
      let val node = hd(path)
          val transChildPairs = Graph.getTransOut(node)
          val newPathsCsList = List.map (fn ((p,comstep),node) => createNewPathCs((comstep,node),P)) transChildPairs
      in List.filter (fn PathCs{statelist,...} => statelist <> []) newPathsCsList end

fun nextPathCsList(pCsList) = Aux.meltList(List.map nextPathCs pCsList)

fun initState(leakterms) = State{labels = [~1],cs=[],scs=[initConstraint(leakterms)],sub=[],psi=[]}

fun mkPathCs(initNode,leakterms) =  PathCs{path=[initNode],statelist = [initState(leakterms)],termList = []}

(* returns SOME(solution) if the state can be solved with the breach otherwise NONE *)
fun solveState(State{labels,cs = [],scs,sub,psi},BreachStates.SEC(t)) = 
      let val (_,IK) = hd(scs) (* final intruder knowledge *)
          val s' = State{labels=labels,cs=[(t,IK)],scs=scs,sub=sub,psi=psi}
          val finals = finalStates [s']
      in if finals = [] then NONE
         else
           let val State{sub,psi,...} = hd(finals) 
               val _ = print'("solution found in solveState SEC from:\n"^stateToString(s'))
           in SOME(sub,psi,hd(finals)) end
      end
  | solveState(s as State{labels,cs,scs,sub,psi},BreachStates.AUT(orig,dest)) = 
      let val _ = print'("searching for breach: ("^cTermToString(orig)^","^cTermToString(dest)^")")
          val _ = print'("in state:\n"^stateToString(s)^"\n")
          val finals = finalStates [s]
      in if finals = [] then NONE
         else
           let val State{sub,psi,...} = hd(finals)
               val _ = if length(finals) <> 1 then print("error in solveState") else ()
           in
             if origDestInPsi((orig,dest),psi)
             then (print' "solution found";SOME(sub,psi,hd(finals)))
             else (print' "no solution\n";NONE)
           end
      end

(* Can this PathCs give us the breach?
we add the finalC to the final states of PathCs and returns SOME solution or NONE *)
fun solution(PathCs{path,statelist=state::states,termList},breach) = 
     (case solveState(state,breach) of
        NONE => solution(PathCs{path=path,statelist=states,termList = termList},breach)
       |SOME(sub,psi,s) => SOME(List.rev path,sub,psi,s))
  | solution(PathCs{statelist=[],...},_) = NONE


fun search(pCs::pCsList,queue,breach) = 
     (case solution(pCs,breach) of
        NONE => search(pCsList,pCs::queue,breach)
       |s as SOME(_) => s)
  | search([],[],breach) = (print "Termination!!";NONE) 
  | search([],queue,breach) = let val next = nextPathCsList(queue)
                                  val l = if List.length(next) > 0 
                                          then let val PathCs{path,...} = hd(next)
                                               in List.length(path) end
                                          else ~1
                                  val _ = print("\nsearching "^Int.toString(List.length(next))^" paths of length "^Int.toString(l)^" \n")
                              in search(next,[],breach) end




end

end

