structure Flow =  
struct

open LysaN_lab
open Types
open Aux
open Graph


fun markFresh nnscope (LysaN.NAME n::ls) = 
      if Aux.member(n,nnscope) 
      then let val (nnscope',ls') = markFresh (Aux.remove(n,nnscope)) ls 
           in (nnscope',LysaN.FRESH(LysaN.NAME n)::ls') end
      else let val (nnscope',ls') = markFresh nnscope ls 
           in (nnscope',LysaN.NAME n::ls') end
  | markFresh nnscope (LysaN.ENC(ts1,key,form)::ts2) = 
      let val (nnscope1',key'::ts1') = markFresh nnscope (key::ts1)
          val (nnscope2',ts2') = markFresh nnscope1' ts2
      in (nnscope2',LysaN.ENC(ts1',key',form)::ts2') end
  | markFresh nnscope (t::ts) = let val (nnscope',ts') = markFresh nnscope ts in (nnscope',t::ts') end
  | markFresh nnscope [] = (nnscope,[])


fun getLabel(OUT(l,_,_)) = SOME(l)
  | getLabel(INP(l,_,_)) = SOME(l)
  | getLabel(NEW(_,P)) = getLabel(P)
  | getLabel(NIL) = NONE
(*  
fun getLast f (Pm' as OUT(_,_,P),Pm) = getLast f (P,Pm')
  | getLast f (Pm' as INP(_,_,P),Pm)= getLast f (P,Pm')
  | getLast f (NEW(_,P),Pm) = getLast f (P,Pm) 
  | getLast f (NIL,Pm) = f(Pm)

fun getLastLabel(P) = case (getLast getLabel (P,P)) of
                       NONE => raise Problem("a process with no labels")
                      |SOME(l) => l
*)
(* this function omits the transition corresponding to a restart *)
fun createFlowP'(OUT(l,ls,P),principal,nnscope,l1) = 
      let val l' = (case (getLabel P) of NONE => l1 | SOME l' => l')  (*either next label of first label if we are at the end *)
          val (nnscope',ls') = markFresh nnscope ls 
      in ((l,l'),(principal,OUTSTEP(ls')))::createFlowP'(P,principal,nnscope',l1) end
  | createFlowP'(INP(l,ls,P),principal,nnscope,l1) = 
      let val l' = (case (getLabel P) of NONE => l1 | SOME l' => l')  (*either next label of first label if we are at the end *)
          val (nnscope',ls') = markFresh nnscope ls 
      in ((l,l'),(principal,INSTEP(ls')))::createFlowP'(P,principal,nnscope',l1) end
  | createFlowP'(NEW(LysaN.NAME n,P),principal,nnscope,l1) =  createFlowP'(P,principal,n::nnscope,l1)
  | createFlowP'(NIL,_,_,_) = []

(*fun getTransType(OUT(l,ls,_)) = OUTSTEP(ls)
  | getTransType(INP(l,ls,_)) = INSTEP(ls)
  | getTransType(NEW _) = raise Problem("getTransType NEW")
  | getTransType(NIL) = raise Problem("getTransType NIL")

fun getLastTransType(P) = getLast getTransType (P,P)
*)
fun createFlowP(P,principal):flow = createFlowP'(P,principal,[],Option.valOf(getLabel(P)))

fun flattenS'(s::Ss) = flattenS s @ flattenS' Ss
  | flattenS'([]) = []

and flattenS(PAR(Ss))  = flattenS' Ss
  | flattenS(LOOP P)  = [LOOP P]
  | flattenS(NEWS(_,S)) = flattenS S 
(*  | flattenS(ANEWS(_,S)) = flattenS S *)


fun extractLeaks(NEW(n,P))    = extractLeaks(P)
  | extractLeaks(OUT(l,ls,P)) = ls@extractLeaks(P)
  | extractLeaks NIL = []
  | extractLeaks _ = raise Problem("")
   
(*fn createFlow: ProcS -> flow *)
fun createFlowS'(LOOP("leak",P)::Ss) = let val (flow,leaks) = createFlowS'(Ss)
                                           val leaks' = extractLeaks(P)
                                       in (flow,leaks@leaks') end
  | createFlowS'(LOOP(n,P)::Ss)  = let val (flow,leaks) = createFlowS'(Ss)
                                       val flow' = createFlowP(P,n)
                                   in (flow'@flow,leaks) end
  | createFlowS'([])             = ([],[])
  | createFlowS'(_)              = raise Problem("createFlowS'")

fun createFlowS(S) = createFlowS'(flattenS(S))

fun getInitState'([]) = []
  | getInitState'(LOOP(_,P)::Ss) = Option.valOf(getLabel(P))::getInitState'(Ss)(*getLastLabel(P)::getInitState'(Ss)*)

fun getInitState(Sl) = let val Sflat = flattenS(Sl)
                           in getInitState'(Sflat) end


(*--------create graph from flow--------------------*)
fun addChildren(nRef,((l1,l2),transition)::flow,queue,graph) = 
             let val state = getState(nRef)
             in 
               (
               case replace(l1,l2,state) of

               SOME(state') => let val (toNodeRef,graph) = addTransition(state,state',transition,graph)
                               in addChildren(nRef,flow,toNodeRef::queue,graph) end
             | NONE => addChildren(nRef,flow,queue,graph)
               )
             end
  | addChildren(nRef,[],queue,graph) = (queue,graph)

fun createGraph'(nRef::queue,visited,flow:flow,graph) = 
          if member(nRef,visited)
          then createGraph'(queue,visited,flow,graph)
          else let val (queue,graph) = addChildren(nRef,flow,queue,graph)
               in createGraph'(queue,nRef::visited,flow,graph) end
  | createGraph'([],_,_,graph) = graph

fun createGraph(S) = let val (flow,leakterms) = createFlowS(S)
                         val nInitRef = mkNode(getInitState(S))
                     in (nInitRef,createGraph'([nInitRef],[],flow,[nInitRef]),leakterms) end

(*----------     collect constraints     ----------------------------------------*)

(*
type path = transType list (* list of inputs and outputs representing a path where DY is involved in every communication*)

type knowledge = LysaN.Term list (* knowle of DY *)
type constraint = LysaN.Term*knowledge


(*  fn: transition * knowledge -> constraint list * knowledge, returns the constraints and updated knowledge of DY corresponding to a transition  *)
fun getRandomPath()=0

fun performTransition(INSTEP(ts)::ls,DY) = let val cs' = List.map (fn t => (t,DY)) ts  (*room for optimization*)
                                                    val (cs,DY) = performTransition(ls,DY)
                                                in (cs'@cs,DY) end
  | performTransition(OUTSTEP(ts)::ls,DY)     = performTransition(ls,Set.union(ts,DY))
  | performTransition([],DY)                  = ([],DY)



fun stateListToNodeList(s::states,nodeRefs) = let val (nodeRef,nodeRefs) = lookUpNodeRef(s,nodeRefs)
                                              in stateListToNodeList(states,nodeRefs) end
                                              
  | stateListToNodeList([],nodeRefs) = nodeRefs
  
*)



(*
Notes

It is more interesting to know wether a transition is involving the attacker than to know wether a state is inside or outside the protocol

*)

end
