(***********************************************************************)
(* Data structure for LysaN.

   Mostly copied from Lysa by Hanne Riis Nielson and Mikael Buchholtz
                                                      Nikolaj H. Kaplan*)
(***********************************************************************)

(*


t ::= n | x | E(t0)(t1,...,tk) 
p ::= n | x | E(p0)(p1,...,pk) | %x | $x

S ::= S|S | (ny n)S | loop(P)

P ::= (ny n)P | <t1,...,tk>.P | (p1,...,pk).P | decrypt t as E(tk)(...)tk | 0

wmf:

PA = (ny K)<A,B,E(KA)(K)>.(ny msg)<E(K)(msg)>.0
PB = (%y).decrypt y as E(KB)(%yK).(%yM).decrypt ym as E(yK)(%msg).0
PC = (ny K)<C,B,E(KA)(K)>.(ny msg)<E(K)(msg)>.0

PS1 = (A,B,%z).decrypt z as E(KA)(%zK).<E(KB)(zK)>.0
PS2 = (C,B,%z).decrypt z as E(KC)(%zK).<E(KB)(zK)>.0



PS' = ($a,$b,%z).decrypt z as E(key[a])(%k).<E(key[b])(k)>.0

Roles:
PA = (ny K)<A,B,E(KA)(K)>.(ny msg)<E(K)(msg)>.0
PB = (%y).decrypt y as E(KB)(%yK).(%yM).decrypt ym as E(yK)(%msg).0
PS = ($A,$B,%z).decrypt z as E(key[A])(%k).<E(key[B])(k)>.0

Principals:
  P1: A
  P2: B
  S:  S
  M:  A,B,S

(ny key[A])(ny key[B])(ny key[M])
Par[
  loop(PA) | loop(PB) | loop(PS)
]




(ny KA)(ny KB)(ny KC)
loop(PA)  |
loop(PB)  |
loop(PC)  |
loop(PS1) | 
loop(PS2) 

*)


(* 

datastruktur til sgning og reduktion
jeg skal bruge en datastruktur, der representerer en protokol specificeret i LySaN. 
Protokellen representeres som en graf / automat som skal representeres af en datastruktur.
Datastrukturen skal have flgende egenskaber:
  1. Det skal vre muligt at sge begge veje i grafen
  2. Det skal vre muligt at fjerne transitioner i grafen
  3. Det skal vre muligt at se p en transition svarende til en kommunikation fra angriber
     og bestemme om den skal fjernes ved at undersge flgende:
       a. Bliver variable i input dekrypteret og kender angriberen dekrypteringsnglen, eller 
       ved at se p en eventuel dekryptering af variable i principalens 
     input-tupel og sammenholde denne dekryptering med angriberens viden. Der skal alts laves en form
     for kobling mellem input-steps og dekrypteringer.

Der skal benyttes en tabel svarende til rho fra analysen

*)






structure LysaN =
struct


type principal = string 
type principals = principal list

type Symbol   = string

type Name     = Symbol
type Var      = Symbol

datatype Form  = ORIG    of Term
               | DEST    of Term

and Term  =      NAME    of Name
               | VAR     of Var
               | DEFVARP of Var
               | DEFVAR  of Var
               | ENC     of Term list * Term * (Term * Form)
               | KEY     of Term (*invariant. Term must be a NAME or a principal var*)
               | FRESH   of Term (*invariant. Term must be a NAME. Not a part of lysaN.*)

datatype ProcP = OUT     of Term list * ProcP
               | INP     of Term list * ProcP 
               | NEW     of Term * ProcP 
               | NIL

datatype ProcS = NEWS    of Term * ProcS
               | PAR     of ProcS list
               | LOOP    of principal * ProcP


type protocol = principals * ProcS

type flatProtocol = ProcP list

end

(************************************************************)

structure LysaN_lab = 
struct

type label = string * int

datatype ProcP = OUT  of label * LysaN.Term list * ProcP
               | INP  of label * LysaN.Term list * ProcP
               | NEW of LysaN.Term * ProcP
               | NIL

datatype ProcS = NEWS of LysaN.Term * ProcS
               | PAR  of ProcS list
               | LOOP of LysaN.principal * ProcP

(* LysaN.ProcS -> LysaN_lab.ProcS *)

val lab = ref 0;
val p = ref "z";
fun nextLabel() = (lab := !lab + 1 ; (!p,!lab))
fun resetLabel(s) = (p := s ; lab := 0)

fun addLabelP(LysaN.OUT(ls,P)) = let val lab = nextLabel() in OUT(lab,ls,addLabelP(P)) end
  | addLabelP(LysaN.INP(ls,P)) = let val lab = nextLabel() in INP(lab,ls,addLabelP(P)) end
  | addLabelP(LysaN.NEW(n,P)) = NEW(n,addLabelP(P)) 
  | addLabelP(LysaN.NIL) = NIL 

fun addLabelS(LysaN.NEWS(n,S)) = NEWS(n,addLabelS(S))
  | addLabelS(LysaN.PAR(ls)) = PAR(List.map addLabelS ls)
  | addLabelS(LysaN.LOOP(name,P)) = (resetLabel(name);LOOP(name,addLabelP(P)))

fun labelToString(p,lab) = p^Int.toString(lab)

end
