structure Perm =
struct

open Types
open Array


(* permutation *)
val n = ref 0
val i = ref 0 (* current selection *)
val l = ref 0 (* length of list *)


(* pow(r,p) = r^p *)
fun pow(r,p) = Real.round(Math.pow(Real.fromInt(r),Real.fromInt(p)))

(* takes an integer and converts to a list of cifers representing a number in radix r. p is the number of ciphers used *)
fun convert(n,r,0) = [n]
  | convert(n,r,p) = let val d = pow(r,p)
                         val c = n div d
                     in if c = 0 then 0::convert(n,r,p-1) else c::convert(n mod d,r,p-1) end

(* takes a list of indices ns and picks out the corresponding elements of ls *)
fun intListToSelection (n::ns,ls) = List.nth(ls,n)::intListToSelection(ns,ls)
  | intListToSelection ([],_) = []


fun enumeratePermutations'(~1,_,_,_) = []
  | enumeratePermutations'(i,r,p,principals) = intListToSelection(convert(i,r,p),principals)::enumeratePermutations'(i-1,r,p,principals)

(* enumerate all possible combinations of n elements selected from the list principals *)
fun enumeratePermutations(principals,n) = let val l = List.length(principals)
                                              val max = pow(l,n)-1
                                          in enumeratePermutations'(max,l,n-1,principals) end




fun enumeratePermutations2'(ls,[]) = List.map (fn x=> [x]) ls
  | enumeratePermutations2'(l::ls,perm) = let val perm' = List.map (fn ls' => l::ls') perm
                                          in perm'@enumeratePermutations2'(ls,perm) end
  | enumeratePermutations2'([],perm) = []

(* takes a list of lists and returns all combinations of the elements: [[1,2],[3,4]] => [[1,3],[1,4],[2,3][2,4]]*)

fun enumeratePermutations2(ls::lss) = let val perm = enumeratePermutations2(lss)
                                      in enumeratePermutations2'(ls,perm) end
  | enumeratePermutations2([]) = []


(* init is called with a 'a list, ls, and a number, n. next returns the next combination of n elements taken from ls. The same element can occur several times *)

fun initPermutation(n',l') = (
  l := l';
  n := n';
  i := pow(!l,!n)-1 (* initial value of i. i is decreased *)
)

fun next(ls) = if List.length(ls) <> !l then raise Problem("list has wrong length") else
               if !i < 0 then (false,[]) else
               let val ns = convert(!i,!l,!n-1)
                   val selection = intListToSelection(ns,ls)
                   val _ = i := !i - 1
               in (true,selection) end


(*

(* selection *)
val pos = ref (array(20,0))
val max = ref 0

fun init(n,max') = let val _ = max := max'
                       val _ = pos := array(n,0)
                       val i = ref 0
                   in while !i < n do (
                     update(!pos,!i,n-(!i));
                     i := !i + 1
                   )
                   end
(*
  public boolean next() {
  	for(int i = 0;i<pos.length;i++) {
  		if(pos[i]<max && (i==0 || pos[i] < pos[i-1]-1)) {
  			int p = ++pos[i];
			if(i != 0) {
				for(int j = i-1;j>=0;j--) {
					pos[j] = ++p;
				}
			}
			return true;
  		}
  	}
  	return false;
  }
*)

fun next() = let val i = ref 0
                 val notFound = ref true
             in (
               while !notFound andalso !i < length(!pos) do (
                 if sub(!pos,!i) < !max andalso (!i = 0 orelse sub(!pos,!i) < sub(!pos,!i-1)-1)
                 then let val _ = update(!pos,!i,sub(!pos,!i)+1)
                          val p = ref (sub(!pos,!i))
                          val _ = notFound := false
                      in if !i <> 0 
                         then let val j = ref (!i - 1)
                              in while !j >= 0 do (
                                   p := (!p + 1);
                                   update(!pos,!j,!p);
                                   j := !j - 1
                                 )
                              end
                         else ()
                      end
                 else ();
                 i := !i + 1
               );
               not (!notFound)
             )
             end

fun nextPerm(ls) = let val i = ref 0
                       val lsPerm = ref []

                   in (
                     while !i < length(!pos) do (
                       lsPerm := List.nth(ls,sub(!pos,!i)-1)::(!lsPerm);
                       i := !i + 1
                     );
                     next();
                     !lsPerm
                   )
                   end

(*                                
  public String toString() {
  	String str = "";
  	int i = 0;
  	for(int j = max; j >= 0 && i < pos.length; j--) {
  		if(pos[i] == j) {
  			str += "x";
  			i++;
  		}
  		else {
  			str += " ";
  		}
  	}
  	return str;
  }
*)

fun out() = let val i = ref 0
                val j = ref (!max)
                val A = ref (array(!max,0))
            in (
              while !j >= 0 andalso !i < length(!pos) do (
                if sub(!pos,!i) = !j then (i := !i + 1;update(!A,!j-1,1)) else ();
                j := !j - 1
              );
              !A
            )
            end

fun printOut() = let val a = out()
                     val i = ref 0
                 in (
                   while !i < length(a) do (
                     print(Int.toString(sub(a,!i)));
                     i := !i +1
                   );
                   print("\n")
                 )
                 end

fun printAll(n,max) = let val _ = init(n,max)
                          val _ = printOut()
                      in while next() do
                        printOut()
                      end


*)
end
