structure ToString=
struct

open Types
open LysaN_lab

val annotations = true

(*aux functions*)
fun quotes(s)= "\""^s^"\""

fun listToString' f (x1::x2::xs) = f(x1)^","^(listToString' f (x2::xs))
  | listToString' f [x] = f(x)
  | listToString' f _ = ""

fun listToStringNL' f (x1::x2::xs) = f(x1)^",\n"^(listToStringNL' f (x2::xs))
  | listToStringNL' f [x] = f(x)
  | listToStringNL' f _ = ""
  
fun listToString f xs = "["^listToString' f xs^"]"

fun listToStringNL f xs = "[\n"^listToStringNL' f xs^"\n]\n"
fun listToStringNL'' f xs = "\n"^listToStringNL' f xs^"\n"


(* latex to string 
fun cpFormToString(cp,form) = 
  if not annotations then "" else 
    case form of 
      LysaN.DEST(t) => "{"^termToString(cp)^"}{\\DEST "^termToString(t)^"}" 
     |LysaN.ORIG(t) => "{"^termToString(cp)^"}{\\ORIG "^termToString(t)^"}"
                                 

and termToString(LysaN.NAME s) = s
  | termToString(LysaN.VAR s) = s
  | termToString(LysaN.DEFVAR s) = "\\%"^s
  | termToString(LysaN.DEFVARP s) = "\\$"^s
  | termToString(LysaN.ENC(ts,key,origdest)) = "\\ENCannot{"^termListToString(ts)^"}{"^termToString(key)^"}"^cpFormToString(origdest)
  | termToString(LysaN.KEY(term)) = "\\KEY{"^termToString(term)^"}"
  | termToString(LysaN.FRESH(term)) = termToString(term)

and termListToString([]) = ""
  | termListToString([t]) = termToString(t)
  | termListToString(t::ts) = termToString(t)^","^termListToString(ts)

fun stepToString(((pName,_),label2),OUTSTEP(ts)) = "[label = \""^pName^"->DY: <"^termListToString(ts)^">\", color = orange]"
  | stepToString(((pName,_),label2),INSTEP(ts)) = "[label = \"DY->"^pName^": ("^termListToString(ts)^")\", color = red]"

fun toStringP(LysaN.OUT(ls,P)) = "\\OUTS{"^termListToString(ls)^"}."^toStringP(P)
  | toStringP(LysaN.INP(ls,P)) = "\\INPS{"^termListToString(ls)^"}."^toStringP(P)
  | toStringP(LysaN.NEW(t,P)) =  "\\NEW{"^termToString(t)^"}"^toStringP(P)
  | toStringP(LysaN.NIL) = "\\RETURN"

fun toStringSs([]) = ""
  | toStringSs(s::ss) = toStringS(s)^"\n|"^toStringSs(ss)

and toStringS(LysaN.PAR(ss)) = listToStringNL'' toStringS ss
  | toStringS(LysaN.LOOP (name,P))  = "["^toStringP(P)^"]\\\\[0ex]"
  | toStringS(LysaN.NEWS(t,S)) = "\\NEW{"^termToString(t)^"}"^toStringS(S)

(*---- cterms ------*)
fun cpFormToStringC(cp,form) = 
  if not annotations then "" else 
    case form of 
      CDEST(t) => "{"^cTermToString(cp)^"}{\\DEST "^cTermToString(t)^"}" 
     |CORIG(t) => "{"^cTermToString(cp)^"}{\\ORIG "^cTermToString(t)^"}"

and cTermToString(CNAME(s,i)) = s^(if i > 0 then "^"^Int.toString i else "")
  | cTermToString(CVAR(s,i)) = s^(if i > 0 then "^"^Int.toString i else "")
  | cTermToString(CENC(ts,key,origdest,b)) = "\\ENCannot{"^cTermListToString(ts)^"}{"^cTermToString(key)^"}"^cpFormToStringC(origdest)^(if b then "*" else "")
  | cTermToString(CKEY(term)) = "\\KEY{"^cTermToString(term)^"}"

and cTermListToString([]) = ""
  | cTermListToString([t]) = cTermToString(t)
  | cTermListToString(t::ts) = cTermToString(t)^","^cTermListToString(ts)

*)
(* ----------- before tex ----------------- *)
fun cpFormToString(cp,form) = 
  if not annotations then "" else 
    case form of 
      LysaN.DEST(t) => "("^termToString(cp)^" DEST "^termToString(t)^")" 
     |LysaN.ORIG(t) => "("^termToString(cp)^" ORIG "^termToString(t)^")"
                                 

and termToString(LysaN.NAME s) = s
  | termToString(LysaN.VAR s) = s
  | termToString(LysaN.DEFVAR s) = "%"^s
  | termToString(LysaN.DEFVARP s) = "$"^s
  | termToString(LysaN.ENC(ts,key,origdest)) = "{"^termListToString(ts)^"}"^termToString(key)^cpFormToString(origdest)
  | termToString(LysaN.KEY(term)) = "Key["^termToString(term)^"]"
  | termToString(LysaN.FRESH(term)) = "fresh["^termToString(term)^"]"

and termListToString([]) = ""
  | termListToString([t]) = termToString(t)
  | termListToString(t::ts) = termToString(t)^","^termListToString(ts)

fun stepToString(((pName,_),label2),OUTSTEP(ts)) = "[label = \""^pName^"->DY: <"^termListToString(ts)^">\", color = orange]"
  | stepToString(((pName,_),label2),INSTEP(ts)) = "[label = \"DY->"^pName^": ("^termListToString(ts)^")\", color = red]"

fun toStringP(LysaN.OUT(ls,P)) = "<"^termListToString(ls)^">."^toStringP(P)
  | toStringP(LysaN.INP(ls,P)) = "("^termListToString(ls)^")."^toStringP(P)
  | toStringP(LysaN.NEW(t,P)) =  "(new "^termToString(t)^")"^toStringP(P)
  | toStringP(LysaN.NIL) = "0"

fun toStringSs([]) = ""
  | toStringSs(s::ss) = toStringS(s)^" | "^toStringSs(ss)

and toStringS(LysaN.PAR(ss)) = listToStringNL toStringS ss
  | toStringS(LysaN.LOOP (name,P))  = toStringP(P)
  | toStringS(LysaN.NEWS(t,S)) = "(new "^termToString(t)^")"^toStringS(S)

(*---- cterms ------*)
fun cpFormToStringC(cp,form) = 
  if not annotations then "" else 
    case form of 
      CDEST(t) => "("^cTermToString(cp)^",DEST "^cTermToString(t)^")" 
     |CORIG(t) => "("^cTermToString(cp)^",ORIG "^cTermToString(t)^")"

and cTermToString(CNAME(s,i)) = s^(if i > 0 then "_"^Int.toString i else "")
  | cTermToString(CVAR(s,i)) = s^(if i > 0 then "_"^Int.toString i else "")
  | cTermToString(CENC(ts,key,origdest,b)) = "{"^cTermListToString(ts)^"}"^cTermToString(key)^cpFormToStringC(origdest)^(if b then "*" else "")
  | cTermToString(CKEY(term)) = "KEY["^cTermToString(term)^"]"

and cTermListToString([]) = ""
  | cTermListToString([t]) = cTermToString(t)
  | cTermListToString(t::ts) = cTermToString(t)^","^cTermListToString(ts)

(*------------------------------------------*)
(*LysaN_lab toString*)

fun toStringPLab(OUT(l,ls,P)) = "<"^termListToString(ls)^">"^labelToString(l)^"."^toStringPLab(P)
  | toStringPLab(INP(l,ls,P)) = "("^termListToString(ls)^")"^labelToString(l)^"."^toStringPLab(P)
  | toStringPLab(NEW(t,P)) =  "("^termToString(t)^")"^toStringPLab(P)
  | toStringPLab(NIL) = "0"

fun toStringSsLab([]) = ""
  | toStringSsLab(s::ss) = toStringSLab(s)^" | "^toStringSsLab(ss)

and toStringSLab(PAR(ss)) = listToString toStringSLab ss
  | toStringSLab(LOOP (name,P))  = toStringPLab(P)
  | toStringSLab(NEWS(t,S)) = "(new "^termToString(t)^")"^toStringSLab(S)



(*LaTex*)
(*
fun stateToString'Tex([]) = ""
  | stateToString'Tex([(p,s)]) = p^Int.toString(s)
  | stateToString'Tex((p,s)::l::state) = p^Int.toString(s)^"."^stateToString'Tex(l::state)

fun stateToStringTex(states:state) = quotes(stateToString'Tex(states))

fun origDestToStringTex(SOME (LysaN.ORIG cps)) = "orig"^listToStringTex (fn x=>x) cps
  | origDestToStringTex(SOME (LysaN.DEST cps)) = "dest"^listToStringTex (fn x=>x) cps
  | origDestToStringTex(NONE) = "NONE"

fun termToStringTex(LysaN.NAME s) = s
  | termToStringTex(LysaN.VAR s) = s
  | termToStringTex(LysaN.DEFVAR s) = "%"^s
  | termToStringTex(LysaN.ENC(ts,key,cp,origdest)) = "{"^termListToStringTex(ts)^"}"^termToStringTex(key)^":"^cp^"<->"^origDestToStringTex(origdest)
  | termToStringTex(LysaN.AENC(ts,key,cp,origdest)) = "{|"^termListToStringTex(ts)^"|}"^termToStringTex(key)^":"^cp^"<->"^origDestToStringTex(origdest)
  | termToStringTex(LysaN.KEY(term)) = "Key["^termToStringTex(term)^"]"

and termListToStringTex([]) = ""
  | termListToStringTex([t]) = termToStringTex(t)
  | termListToStringTex(t::ts) = termToStringTex(t)^","^termListToStringTex(ts)

fun stepToStringTex(((pName,_),label2),OUTSTEP(ts)) = "[label = \""^pName^"->DY: <"^termListToStringTex(ts)^">\", color = orange]"
  | stepToStringTex(((pName,_),label2),INSTEP(ts)) = "[label = \"DY->"^pName^": ("^termListToStringTex(ts)^")\", color = red]"

fun toStringPTex(OUT(l,ls,P)) = "<"^termListToStringTex(ls)^">"^labelToStringTex(l)^"."^toStringPTex(P)
  | toStringPTex(INP(l,ls,P)) = "("^termListToStringTex(ls)^")"^labelToStringTex(l)^"."^toStringPTex(P)
  | toStringPTex(NEWP(t,P)) = "(ny "^termToStringTex(t)^")"^toStringPTex(P)
  | toStringPTex(ANEWP(t,P)) = "(ny +- "^termToStringTex(t)^")"^toStringPTex(P)
  | toStringPTex(NIL) = "0"

fun toStringSsTex([]) = ""
  | toStringSsTex(s::ss) = toStringSTex(s)^" | "^toStringSsTex(ss)

and toStringSTex(PAR(ss)) = toStringSsTex(ss)
  | toStringSTex(LOOP (name,P))  = toStringPTex(P)
  | toStringSTex(NEWS(t,S)) = "(ny "^termToStringTex(t)^")"^toStringSTex(S)
  | toStringSTex(ANEWS(t,S)) = "(ny +- "^termToStringTex(t)^")"^toStringSTex(S)


fun stepToStringTex((l1,l2),OUTSTEP(ts)) = "\\leftangle"^termListToStringTex(ts)^"\\rightangle"
  | stepToStringTex((l1,l2),INSTEP(ts)) = "("^termListToStringTex(ts)^")"
*)

(*lysa to string*)

fun symbolToString(s,is) = s (*ignores indices*)

fun origDestToStringLysa(SOME (Lysa.ORIG cps)) = " ORIG "^listToString symbolToString cps
  | origDestToStringLysa(SOME (Lysa.DEST cps)) = " DEST "^listToString symbolToString cps
  | origDestToStringLysa(NONE) = "NONE"

fun termToStringLysa(Lysa.NAME s) = symbolToString s
  | termToStringLysa(Lysa.VAR s) = symbolToString s
(*  | termToString(Lysa.ENC(ts,key,cp,origdest)) = "{"^termListToString(ts)^"}"^termToString(key) *)
(*  | termToString(Lysa.AENC(ts,key,cp,origdest)) = "{|"^termListToString(ts)^"|}"^termToString(key) *)
  | termToStringLysa(Lysa.ENC(ts,key,cp,origdest)) =
        "{"^termListToStringLysa(ts)^"}"^termToStringLysa(key)^"("^(symbolToString cp)^origDestToStringLysa(origdest)^")"
  | termToStringLysa(Lysa.AENC(ts,key,cp,origdest)) =
        "{|"^termListToStringLysa(ts)^"|}"^termToStringLysa(key)^"("^(symbolToString cp)^origDestToStringLysa(origdest)^")"

and termListToStringLysa([]) = ""
  | termListToStringLysa([t]) = termToStringLysa(t)
  | termListToStringLysa(t::ts) = termToStringLysa(t)^","^termListToStringLysa(ts)


fun toStringLysa(Lysa.OUT(ls,P)) = "<"^termListToStringLysa(ls)^">."^toStringLysa(P)
  | toStringLysa(Lysa.INP(ts,vs,P)) = "("^termListToStringLysa(ts)^";"^termListToStringLysa(vs)^")."^toStringLysa(P)
  | toStringLysa(Lysa.NEW(t,P)) = "(ny "^termToStringLysa(t)^")"^toStringLysa(P)
  | toStringLysa(Lysa.ANEW(t,P)) = "(ny +- "^termToStringLysa(t)^")"^toStringLysa(P)
  | toStringLysa(Lysa.PAR(Ps)) = listToStringNL toStringLysa Ps
  | toStringLysa(Lysa.BANG(P)) = "! "^toStringLysa(P)
  | toStringLysa(Lysa.DEC(t,ts,vs,key,cp,origdest,P)) = "decrypt "^termToStringLysa(t)^" as {"^termListToStringLysa(ts)^";"^termListToStringLysa(vs)^"}"^
                                                      termToStringLysa(key)^"("^(symbolToString cp)^origDestToStringLysa(origdest)^")"^"."^toStringLysa(P)
  | toStringLysa(Lysa.NIL) = "0"

(*graph toString functions *)

fun stateToString'([]) = ""
  | stateToString'([(p,s)]) = p^Int.toString(s)
  | stateToString'((p,s)::l::state) = p^Int.toString(s)^"."^stateToString'(l::state)

fun stateToString(states:state) = quotes(stateToString'(states))

fun stepToStringGraph(((pName,_),label2),OUTSTEP(ts)) = "[label = \""^pName^"->DY: <"^termListToString(ts)^">\", color = orange]"
  | stepToStringGraph(((pName,_),label2),INSTEP(ts)) = "[label = \"DY->"^pName^": ("^termListToString(ts)^")\", color = red]"

fun stepToString((l1,l2),OUTSTEP(ts)) = "<"^termListToString(ts)^">"
  | stepToString((l1,l2),INSTEP(ts)) = "("^termListToString(ts)^")"


  


end