structure UI =
struct

(*
    Userinterface for analysis and modelchecking. The interactive program is started by calling the function
    startTool() with no arguments.
    User is asked to choose a protocol from a list. For instance:
      1. Wide mouthed frog
      2. ...
    Choose a number:
    The LySaN representation of the protocol is printed to the screen
    User is asked to type a set of principals such as: A,B,I
    The LySa representation is printed to the screen
    User is asked to confirm or restart
    Confirm

*)

local
open TextIO
open List
open Analysis
open ToString
open Types

in
                                 
val tokenfun = (fn x => x = #"(" orelse x = #")" orelse x = #"," orelse x = #"\n" orelse x = #"[" orelse x = #"]")

fun printProtocolList'(i,(name,_)::ls) = (print(Int.toString(i)^". "^name^"\n");printProtocolList'(i+1,ls))
  | printProtocolList'(_,[]) = ()
  
fun printProtocolList() = printProtocolList'(1,Prot.protocolList)

fun inputProtocolNumber() = let val input = Option.valOf(Int.fromString(inputLine(stdIn)))
                            in List.nth(Prot.protocolList,input-1) end

fun inputPrincipalSet() =  String.tokens tokenfun (inputLine(stdIn))

fun lastChar(str) = String.extract(str,size str - 1,NONE)

fun indexOf'(x,x'::xs,i) = if x = x' then i else indexOf'(x,xs,i+1)
  | indexOf'(_,[],i) = ~1
fun indexOf(x,xs) = indexOf'(x,xs,0)

fun parseName(str) = let val xs = String.explode(str)
                     in case (indexOf(#"_",xs)) of
                       ~1 => CNAME(str,0)
                      |i  => (case Int.fromString(lastChar str) of
                                SOME(index) => CNAME(String.substring(str,0,size str -2),index) (*assuming that the index is one digit!*)
                               |NONE => raise Problem("")   
                             )
                     end     

fun inputBreach() = let val input = String.tokens tokenfun (inputLine(stdIn))
                    in if hd(input) = "KEY" then BreachStates.SEC(CKEY(CNAME(hd(tl input),0))) else
                    let val input = map parseName input
                    in 
                      case List.length(input) of
                        1 => BreachStates.SEC(hd(input))
                       |2 => BreachStates.AUT(hd(input),hd(tl(input)))
                    end
                    end

fun inputTermList() = let val input = String.tokens tokenfun (inputLine(stdIn))
                      in map parseName input end


fun startTool() = let val _ = printProtocolList()
                      val _ = print("Choose a protocol by entering number:\n")
                      val (name,protocol) = inputProtocolNumber()
                      val _ = print("LySaN specification of "^name^"\n"^ToString.toStringS(protocol))
                      val _ = print("Choose a set of principals:\n")
                      val principals = inputPrincipalSet()
                      val protocolLysa = LysaN2Lysa.convertS(protocol,principals)
                      val _ = print("LySa specification of "^name^"\n"^ToString.toStringLysa(protocolLysa))
                      val solution = Analysis.analyzeLysa(protocolLysa)
                      val unf = unfoldLDY(solution,0)
                      val _ = print(psiToString(solution))
                      val _ = print("kappa = \n")
                      val _ = print((ToString.listToString ntToString unf)^"\n")
(*                    val _ = print("specify terms to be leaked\n")
                      val leaks = inputTermList()
*)                    val _ = print("type a breach:\n")
                      val breach = inputBreach()
                      val (path,sub,psi,state) = MC.searchForBreach(breach,protocol)
(*                      val _ = print("path:\n"^Graph.pathToString(path)^"\n")*)
(*                      val narration = Graph.createNarration(path,[],sub)*)
                      val narration = Graph.createNarration(path,[],fixSubstitution(sub))
                      val _ = print("\nnarration:\n"^Graph.narrationToString(narration)^"\n")
                      val _ = print("final "^EvalConstraints.stateToString(state)^"\n") 
(*                      val _ = print("psi "^Constraints.psiToString(psi)^"\n") *)
                   in () end




(*
fun inputAuthenticationBreach(solution) = let val _ = print "input pair of cryptopoints\n"
                                              val input = String.tokens tokenfun (inputLine(stdIn))
                                              val input = map (fn x => LysaN.NAME(x)) input
                                          in if List.length(input) <> 2 
                                             then error(solution)
                                             else (print("("^ToString.termToString(nth(input,0))^","^ToString.termToString(nth(input,1))^")\n");inputCommand(solution)) end
                               
and inputNT(solution) = let val _ = print "input nonterminal\n"
                             val nt = inputLine(stdIn)
                             val nt = NameNT(String.substring(nt,0,String.size(nt)-1))
                             val _ = print(Analysis.ntsToString(Analysis.unfold solution 1 nt)^"\n")
                         in inputCommand(solution) end

and inputCommand(solution) = let val _ = print "input command\n"
                                 val cmd = inputLine stdIn
                             in case cmd of
                               "unfold\n" => inputNT(solution)
                              |"autBreach\n" => (inputAuthenticationBreach(solution);())
                              |"quit\n" => ()
                              | _ => error(solution)
                             end
and error(solution) = let val _ = print "error in input\n"
                      in inputCommand(solution) end


*)
end
end